#!/usr/bin/env python3
#
# Copyright 2022 Max Planck Insitute Magdeburg
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
#
#
"""Container for strain design solutions (SDSolutions)"""

from numpy import all, any, nan, isnan, sign
from typing import List, Dict, Tuple, Union, Set, FrozenSet
from straindesign.parse_constr import *
from straindesign.names import *
import re
import json
import pickle
import logging


class SDSolutions(object):
    """Container for strain design solutions
    
    Objects of this class are returned by strain design computations. This class
    contains the metabolic interventions on the gene, reaction or regulation level
    alongside with information about the strain design setup, including the model
    used and the strain design modules. Strain design solutions can be accessed
    either through the fields or through specific functions that preprocess or
    reformat strain designs for different purposes.
    
    Instances of this class are not meant to be created by StrainDesign users.
    
    Args:
        model (cobra.Model):
            A metabolic model that is an instance of the cobra.Model class.
            
        sd (list of dict):
            A list of dicts every dict represents an intervention set. Keys in
            each dict are reaction/gene identifiers and the associated value
            determines if it is added (1), not added (0) or knocked out (-1).
            For regulatory interventions, (1) means active regulation and 
            (0) means regulatory intervention not added. These will be translated
            to True and False.
        
        status (str):
            Status string of the computation (e.g.: 'optimal')
            
        sd_setup (dict):
            A dictionary containing information about the problem setup. This dict can/should contain
            the keys MODEL_ID, MODULES, MAX_SOLUTIONS, MAX_COST, TIME_LIMIT, SOLVER, KOCOST, KICOST, 
            REGCOST, GKICOST, GKOCOST
            
            These entries can be set like this:
            sd_setup[straindesign.MODEL_ID] = model.id
            
    Returns
        (SDSolutions):
        Strain design solutions
        
    """

    def __init__(self, model, sd, status, sd_setup):
        self.status = status
        self.sd_setup = sd_setup
        if GKOCOST in sd_setup or GKICOST in sd_setup:
            self.gene_sd = [s.copy() for s in sd]
            self.is_gene_sd = True
            logging.info('  Preparing (reaction-)phenotype prediction of gene intervention strategies.')
            interventions = set()
            [[interventions.add(k) for k in s.keys()] for s in sd]
            # replace gene names with identifiers if necessary
            gene_name_id_dict = {}
            [gene_name_id_dict.update({g.name: g.id}) for g in model.genes if interventions.intersection([g.name])]
            for g_name, g_id in gene_name_id_dict.items():
                for s in sd:
                    if g_name in s:
                        s.update({g_id: s.pop(g_name)})
                interventions.remove(g_name)
                interventions.add(g_id)
            # get potential gene- and reaction interventions and potentially affected reactions
            reac_itv = set(v for v in interventions if model.reactions.has_id(v))
            gene_itv = set(v for v in interventions if v in model.genes.list_attr('name') + model.genes.list_attr('id'))
            regl_itv = set(v for v in interventions if v not in reac_itv and v not in gene_itv)
            affected_reacs = set()
            [[affected_reacs.add(r.id) for r in g.reactions] for g in model.genes]
            gpr = {}
            for r in affected_reacs:
                gpr_i = model.reactions.get_by_id(r).gene_reaction_rule
                cj_terms = gpr_i.split(' or ')
                for i, c in enumerate(cj_terms):
                    cj_terms[i] = c.split(' and ')
                    for j, g in enumerate(cj_terms[i]):
                        cj_terms[i][j] = re.sub(r'^(\s|-|\+|\()*|(\s|-|\+|\))*$', '', g)
                gpr.update({r: cj_terms})

            # # get gpr rules
            # gpr = {r.id:r.gene_reaction_rule for r in model.reactions if r.id in affected_reacs}
            # [gpr.update({k:parse_expr(v.replace(' or ',' | ').replace(' and ',' & '))}) for k,v in gpr.items()]
            # translate every cut set to reaction intervention sets
            self.reaction_sd = [{} for _ in range(len(sd))]
            for i, s in enumerate(sd):
                reac_ko = set(k for k, v in s.items() if v < 0 and (k in reac_itv))
                reac_ki = set(k for k, v in s.items() if v > 0 and (k in reac_itv))
                reac_no_ki = set(k for k, v in s.items() if v == 0 and (k in reac_itv))
                reg_itv = set(k for k, v in s.items() if v and (k in regl_itv))
                reg_no_itv = set(k for k, v in s.items() if not v and (k in regl_itv))
                gene_ko = {k: False for k, v in s.items() if v < 0 and (k in gene_itv)}
                gene_ki = {k: True for k, v in s.items() if v > 0 and (k in gene_itv)}
                gene_no_ki = {k: False for k, v in s.items() if v == 0 and (k in gene_itv)}
                gene_ki_inv = {k: False for k in gene_ki.keys()}
                for r in affected_reacs:
                    # is_possible = gpr[r].subs({**gene_ko,**gene_ki,**gene_no_ki})
                    is_possible = gpr_eval(gpr[r], {**gene_ko, **gene_ki, **gene_no_ki})
                    if is_possible != False:  # reaction is only feasible because of KIs
                        # is_possible_wo_ki = gpr[r].subs({**gene_ko,**gene_ki_inv,**gene_no_ki})
                        is_possible_wo_ki = gpr_eval(gpr[r], {**gene_ko, **gene_ki_inv, **gene_no_ki})
                        if is_possible_wo_ki == False:
                            reac_ki.add(r)
                    elif is_possible == False:
                        # is_possible_wo_ko = gpr[r].subs({**gene_ki,**gene_no_ki})
                        is_possible_wo_ko = gpr_eval(gpr[r], {**gene_ki, **gene_no_ki})
                        if is_possible_wo_ko != False:
                            reac_ko.add(r)
                        else:
                            reac_no_ki.add(r)
                    else:  # reaction is not (sufficiently) affected by interventions
                        pass
                self.reaction_sd[i].update({k: -1.0 for k in reac_ko})
                self.reaction_sd[i].update({k: 1.0 for k in reac_ki})
                self.reaction_sd[i].update({k: 0.0 for k in reac_no_ki})
                self.reaction_sd[i].update({k: True for k in reg_itv})
                self.reaction_sd[i].update({k: False for k in reg_no_itv})
        else:
            self.reaction_sd = sd
            self.is_gene_sd = False

        if GKOCOST in sd_setup or GKICOST in sd_setup:
            sd = [s.copy() for s in self.gene_sd]

        # compute intervention costs
        self.sd_cost = [0 for _ in range(len(sd))]
        if KOCOST in sd_setup:
            for k, v in sd_setup[KOCOST].items():
                for i, s in enumerate(sd):
                    if k in s and s[k] != 0:
                        self.sd_cost[i] += float(v)
        if KICOST in sd_setup:
            for k, v in sd_setup[KICOST].items():
                for i, s in enumerate(sd):
                    if k in s and s[k] != 0:
                        self.sd_cost[i] += float(v)
        if GKOCOST in sd_setup:
            for k, v in sd_setup[GKOCOST].items():
                for i, s in enumerate(sd):
                    if k in s and s[k] != 0:
                        self.sd_cost[i] += float(v)
        if GKICOST in sd_setup:
            for k, v in sd_setup[GKICOST].items():
                for i, s in enumerate(sd):
                    if k in s and s[k] != 0:
                        self.sd_cost[i] += float(v)
        if REGCOST in sd_setup:
            for k, v in sd_setup[REGCOST].items():
                for i, s in enumerate(sd):
                    if k in s and s[k] != 0:
                        self.sd_cost[i] += float(v)

        self.has_complex_regul_itv = False
        self.itv_bounds = [{} for _ in range(len(self.reaction_sd))]
        for i, s in enumerate(self.reaction_sd):
            for k, v in s.items():
                if type(v) is not bool:
                    if v == -1:  # reaction was knocked out
                        self.itv_bounds[i].update({k: (0.0, 0.0)})
                    elif v == 1:  # reaction was added
                        self.itv_bounds[i].update({k: model.reactions.get_by_id(k).bounds})
                    elif v == 0:  # reaction was not added
                        self.itv_bounds[i].update({k: (nan, nan)})
            for k, v in s.items():
                if type(v) == bool and v:
                    try:
                        lineq = lineq2list([k], model.reactions.list_attr('id'))[0]
                    except:
                        self.has_complex_regul_itv = True
                        continue
                    lhs = lineq[0]
                    if len(lhs) != 1:
                        self.has_complex_regul_itv = True
                    else:
                        eqsign = lineq[1]
                        rhs = lineq[2]
                        reac = list(lhs.keys())[0]
                        coeff = list(lhs.values())[0]
                        if reac in self.itv_bounds[i]:
                            bnds = self.itv_bounds[i].pop(reac)
                        else:
                            bnds = model.reactions.get_by_id(reac).bounds
                        if eqsign == '=':
                            bnds = (rhs / coeff, rhs / coeff)
                        elif (eqsign == '<=') == (sign(coeff) > 0):
                            bnds = (bnds[0], rhs / coeff)
                        else:
                            bnds = (rhs / coeff, bnds[1])
                        self.itv_bounds[i].update({reac: bnds})

    def get_num_sols(self):
        """Get number of solutions"""
        return len(self.reaction_sd)

    def get_strain_design_costs(self, i=None):
        """Get costs of i-th strain design or of all in a list"""
        if i is None:
            return self.sd_cost
        else:
            return get_subset(self.sd_cost, i)

    def get_strain_designs(self, i=None):
        """Get i-th strain design (intervention set) or all in original format"""
        if self.is_gene_sd:
            return self.get_gene_sd(i)
        else:
            return self.get_reaction_sd(i)

    def get_reaction_sd(self, i=None):
        """Get reaction-based strain design solutions
        
        Gene-based intervention sets are translated to the reaction level. This can
        be helpful to understand the impact of gene interventions. GPR-rules are 
        accounted for automatically.
        """
        if i is None:
            return [strip_non_ki(s) for s in self.reaction_sd]
        else:
            if type(i) == int:
                i = [i]
            return [strip_non_ki(s) for j, s in enumerate(self.reaction_sd) if j in i]

    def get_reaction_sd_bnds(self, i=None):
        """Get reaction-based strain design solutions represented by upper and lower bounds
        
        Knocked-out reactions will show as upper and lower bounds of zero.
        """
        if i is None:
            return self.itv_bounds
        else:
            if type(i) == int:
                i = [i]
            return [self.itv_bounds for j, s in enumerate(self.itv_bounds) if j in i]

    def get_gene_sd(self, i=None):
        """Get gene-based strain design solutions"""
        if not self.is_gene_sd:
            raise Exception('The solutions are based on reaction interventions only.')
        if i is None:
            return [strip_non_ki(s) for s in self.gene_sd]
        else:
            if type(i) == int:
                i = [i]
            return [strip_non_ki(s) for j, s in enumerate(self.gene_sd) if j in i]

    def get_reaction_sd_mark_no_ki(self, i=None):
        """Get reaction-based strain design solutions, 
        but also tag knock-ins that were not made with a 0
        
        This can be helpful to analyze gene intervention sets in original metabolic models.
        GPR-rules are accounted for automatically."""
        if i is None:
            return self.reaction_sd
        else:
            if type(i) == int:
                i = [i]
            return get_subset(self.reaction_sd, i)

    def get_gene_sd_mark_no_ki(self, i=None):
        """Get gene-based strain design solutions, 
        but also tag knock-ins that were not made with a 0"""
        if not self.is_gene_sd:
            raise Exception('The solutions are based on reaction interventions only.')
        if i is None:
            return self.gene_sd
        else:
            if type(i) == int:
                i = [i]
            return get_subset(self.gene_sd, i)

    def get_gene_reac_sd_assoc(self, i=None):
        """Get reaction and gene-based strain design solutions, 
        and show which reaction-based solution corresponds to which gene-based.
        
        Often the association is not 1:1 but n:1."""
        if not self.is_gene_sd:
            raise Exception('The solutions are based on reaction interventions only.')
        if i is None:
            i = [j for j in range(len(self.gene_sd))]
        else:
            if type(i) == int:
                i = [i]
        reacs_sd_hash = []
        reacs_sd = []
        assoc = []
        gene_sd = [strip_non_ki(s) for j, s in enumerate(self.gene_sd) if j in i]
        for s in [strip_non_ki(t) for j, t in enumerate(self.reaction_sd) if j in i]:
            hs = hash(json.dumps(s, sort_keys=True))
            if hs not in reacs_sd_hash:
                reacs_sd_hash.append(hs)
                reacs_sd.append(s)
            assoc.append(reacs_sd_hash.index(hs))
        return reacs_sd, assoc, gene_sd

    def get_gene_reac_sd_assoc_mark_no_ki(self, i=None):
        """Get reaction and gene-based strain design solutions, 
        but also tag knock-ins that were not made with a 0
        
        Often the association is not 1:1 but n:1."""
        if not self.is_gene_sd:
            raise Exception('The solutions are based on reaction interventions only.')
        if i is None:
            i = [j for j in range(len(self.gene_sd))]
        else:
            if type(i) == int:
                i = [i]
        reacs_sd_hash = []
        reacs_sd = []
        assoc = []
        gene_sd = [s for j, s in enumerate(self.gene_sd) if j in i]
        for s in [t for j, t in enumerate(self.reaction_sd) if j in i]:
            hs = hash(json.dumps(s, sort_keys=True))
            if hs not in reacs_sd_hash:
                reacs_sd_hash.append(hs)
                reacs_sd.append(s)
            assoc.append(reacs_sd_hash.index(hs))
        return reacs_sd, assoc, gene_sd

    def save(self, filename):
        """Save strain design solutions to a file."""
        with open(filename, 'wb') as f:
            pickle.dump(self, f)

    @classmethod
    def load(cls, filename):
        """Load strain design solutions from a file."""
        with open(filename, 'rb') as f:
            cls = pickle.load(f)
        return cls


def strip_non_ki(sd):
    """SDSolutions internal function: removing non-added reactions or genes"""
    return {k: v for k, v in sd.items() if v not in (0.0, False)}


def get_subset(sd, i):
    """SDSolutions internal function: getting a subset of solutions"""
    return [s for j, s in enumerate(sd) if j in i]


def gpr_eval(cj_terms, interv):
    """SDSolutions internal function: evaluate a GPR term"""
    gpr_ev = [0.0 for _ in range(len(cj_terms))]
    for i, c in enumerate(cj_terms):
        cj_term = [interv[k] if k in interv else nan for k in c]
        if any([v == False for v in cj_term]):
            gpr_ev[i] = False
        elif any(isnan(cj_term)):
            gpr_ev[i] = nan
        elif all(cj_term):
            gpr_ev[i] = True

    if any([v == True for v in gpr_ev]):
        return True
    elif all([v == False for v in gpr_ev]):
        return False
    elif any(isnan(gpr_ev)):
        return nan
    else:
        raise Exception('Shoud not happen')



#!/usr/bin/env python3
#
# Copyright 2022 Max Planck Insitute Magdeburg
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
#
"""Class for indicator contraints (IndicatorConstraints)"""


class IndicatorConstraints:
    """A class for storing indicator contraints
    
    This class is a container for indicator constraints. Indicator constraints are used to link the fulfillment
    of a constraint to an indicating variable. For instance the indicator constraint:    
    z = 1 -> 2*d - 1*e <= 3
    can be described as: 
    If z=1, then 2*d - 1*e <= 3
    An alternative formulation of this association is possible with a bigM constraint:
    2*d - 1*e - z*M <= 3, with M = very large
    The constraint z = 0 -> 2*d - 1*e <= 3 would translate to 2*d - 1*e + z*M <= 3 + M
    Generally, indicator constraints are preferred over bigM, because they provide better numerical stability.
    However, not all solvers support indicator constraints

    Indicator constraints have the form:
    x_binv = indicval -> a * x <sense> b
    e.g.,: x_35 = 1 -> 2 * x_2 + 3 *x_3 'L' 6
                                        (<=)
    This class contains a set of indicator constraints:
    x_binv_1 = indicval_1 -> A_1 * x <sense_1> b_1
    x_binv_2 = indicval_2 -> A_2 * x <sense_2> b_2
    ...

    Example: 
        ic = IndicatorConstraints(binv, A, b, sense, indicval)
                    
    Args:
        binv (list of int): (e.g.: [25, 27, 30])
            The index of the binary, indicating variables (indicators) for all indicator constraints.
            Integers are allowed ot occur more than once.
        
        A (sparse.csr_matrix):
            Coefficient vectors for all indicator constraints, stored in one matrix, whereas
            each row is used in one indicator constraint. 
            (num_columns = number of variables, num_rows = number of indicator constraints)

        b (list of float):
            Right hand sides of all indicator constraints. (e.g.: [0.1, 2.0, 3])
            
        sense (str):
            (In)equality signs for all indicator constraints. 
            'L'ess or equal, 'E'qual or, 'G'reater or equal (e.g.: 'EEGLGLGE')
            
        indicval (list of int):
            Indicator values for all indicator constraints. Which value of the indicator 
            enforces the constraint, 0 or 1? (e.g., 0001010110)
            
    Returns:
        (IndicatorConstraints):
        An object of the IndicatorConstraints class to pass indicator constraints.
    """

    def __init__(self, binv, A, b, sense, indicval):
        self.binv = binv  # index of binary variable
        self.A = A  # CPLEX: lin_expr,   left hand side coefficient row for indicator constraint
        self.b = b  # right hand side for indicator constraint
        self.sense = sense  # sense of the indicator constraint can be 'L', 'E', 'G' (lower-equal, equal, greater-equal)
        self.indicval = indicval  # value the binary variable takes when constraint is fulfilled



#!/usr/bin/env python3
#
# Copyright 2022 Max Planck Insitute Magdeburg
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
#
"""CPLEX solver interface for LP and MILP"""

from scipy import sparse
from numpy import nan, inf, isinf, random
from cplex import Cplex, infinity, _const
from cplex.exceptions import CplexError
from typing import Tuple, List
import logging
import io
from psutil import virtual_memory
from straindesign.names import *


class Cplex_MILP_LP(Cplex):
    """CPLEX interface for MILP and LP
    
    This class is a wrapper for the CPLEX-Python API to offer bindings and namings
    for functions for the construction and manipulation of MILPs and LPs in an
    vector-matrix-based manner that are consistent with those of the other solver 
    interfaces in the StrainDesign package. The purpose is to unify the instructions 
    for operating with MILPs and LPs throughout StrainDesign.
    
    The CPLEX interface provides support for indicator constraints as well as for
    the populate function.
    
    Accepts a (mixed integer) linear problem in the form:
        minimize(c),
        subject to: 
        A_ineq * x <= b_ineq,
        A_eq * x  = b_eq,
        lb <= x <= ub,
        forall(i) type(x_i) = vtype(i) (continous, binary, integer),
        indicator constraints:
        x(j) = [0|1] -> a_indic * x [<=|=|>=] b_indic
                    
    Please ensure that the number of variables and (in)equalities is consistent
        
    Example: 
        cplex = Cplex_MILP_LP(c, A_ineq, b_ineq, A_eq, b_eq, lb, ub, vtype, indic_constr)
                
    Args:
        c (list of float): (Default: None)
            The objective vector (Objective sense: minimization).
            
        A_ineq (sparse.csr_matrix): (Default: None)
            A coefficient matrix of the static inequalities.   
            
        b_ineq (list of float): (Default: None)
            The right hand side of the static inequalities.
            
        A_eq (sparse.csr_matrix): (Default: None)
            A coefficient matrix of the static equalities.   
            
        b_eq (list of float): (Default: None)
            The right hand side of the static equalities.
            
        lb (list of float): (Default: None)
            The lower variable bounds.
            
        ub (list of float): (Default: None)
            The upper variable bounds.
            
        vtype (str): (Default: None)
            A character string that specifies the type of each variable:
            'c'ontinous, 'b'inary or 'i'nteger
            
        indic_constr (IndicatorConstraints): (Default: None)
            A set of indicator constraints stored in an object of IndicatorConstraints
            (see reference manual or docstring).
        
        seed (int16): (Default: None)
            An integer value serving as a seed to make MILP solving reproducible.
        
        Returns:
            (Cplex_MILP_LP):
            
            A CPLEX MILP/LP interface class.
    """

    def __init__(self, c=None, A_ineq=None, b_ineq=None, A_eq=None, b_eq=None, lb=None, ub=None, vtype=None, indic_constr=None, seed=None):
        super().__init__()
        self.objective.set_sense(self.objective.sense.minimize)
        try:
            numvars = A_ineq.shape[1]
        except:
            numvars = A_eq.shape[1]
        # replace numpy inf with cplex infinity
        for i, v in enumerate(b_ineq):
            if isinf(v):
                b_ineq[i] = infinity
        # concatenate right hand sides
        b = b_ineq + b_eq
        # prepare coefficient matrix
        if isinstance(A_eq, list):
            if not A_eq:
                A_eq = sparse.csr_matrix((0, numvars))
        if isinstance(A_ineq, list):
            if not A_ineq:
                A_ineq = sparse.csr_matrix((0, numvars))
        A = sparse.vstack((A_ineq, A_eq), format='coo')  # concatenate coefficient matrices
        sense = len(b_ineq) * 'L' + len(b_eq) * 'E'

        # construct CPLEX problem. Add variables and linear constraints
        self.variables.add(obj=c, lb=lb, ub=ub, types=vtype)
        self.linear_constraints.add(rhs=b, senses=sense)
        if A.nnz:
            self.linear_constraints.set_coefficients(zip(A.row.tolist(), A.col.tolist(), A.data.tolist()))

        # add indicator constraints
        if not indic_constr == None:
            # cast variables and translate coefficient matrix A to right input format for CPLEX
            A = [[[int(i) for i in list(a.indices)], [float(i) for i in list(a.data)]] for a in sparse.csr_matrix(indic_constr.A)]
            b = [float(i) for i in indic_constr.b]
            sense = [str(i) for i in indic_constr.sense]
            indvar = [int(i) for i in indic_constr.binv]
            complem = [1 - int(i) for i in indic_constr.indicval]
            # call CPLEX function to add indicators
            self.indicator_constraints.add_batch(lin_expr=A, sense=sense, rhs=b, indvar=indvar, complemented=complem)
        # set parameters
        self.set_log_stream(io.StringIO())  # don't show output stream
        self.set_error_stream(io.StringIO())
        self.set_warning_stream(io.StringIO())
        self.set_results_stream(io.StringIO())
        self.parameters.simplex.tolerances.optimality.set(1e-9)
        self.parameters.simplex.tolerances.feasibility.set(1e-9)

        if 'B' in vtype or 'I' in vtype:
            # set usable working memory to 3/4 of the total available memory
            self.parameters.workmem.set(round(virtual_memory().total / 1024 / 1024 * 0.75))
            #self.parameters.threads.set(16)
            # yield only optimal solutions in pool
            if seed is None:
                # seed = random.randint(0, _const.CPX_BIGINT)
                seed = int(random.randint(0, 2**16 - 1))
            self.parameters.randomseed.set(seed)
            self.parameters.mip.pool.absgap.set(0.0)
            self.parameters.mip.pool.relgap.set(0.0)
            self.parameters.mip.pool.intensity.set(4)
            # no integrality tolerance
            self.parameters.mip.tolerances.integrality.set(0.0)

    def solve(self) -> Tuple[List, float, float]:
        """Solve the MILP or LP
        
        Example:
            sol_x, optim, status = cplex.solve()
        
        Returns:
            (Tuple[List, float, float])
            
            solution_vector, optimal_value, optimization_status
        """
        try:
            super().solve()  # call parent solve function (that was overwritten in this class)
            status = self.solution.get_status()
            if status in [1, 101, 102, 115, 128, 129, 130]:  # solution integer optimal
                min_cx = self.solution.get_objective_value()
                status = OPTIMAL
            elif status == 108:  # timeout without solution
                x = [nan] * self.variables.get_num()
                min_cx = nan
                status = TIME_LIMIT
                return x, min_cx, status
            elif status in [3, 103]:  # infeasible
                x = [nan] * self.variables.get_num()
                min_cx = nan
                status = INFEASIBLE
                return x, min_cx, status
            elif status in [11, 107]:  # timeout with solution
                min_cx = self.solution.get_objective_value()
                status = TIME_LIMIT_W_SOL
            elif status in [2, 4, 118, 119]:  # solution unbounded
                x = [nan] * self.variables.get_num()
                min_cx = -inf
                status = UNBOUNDED
                return x, min_cx, status
            else:
                logging.exception(status)
                logging.exception(self.solution.get_status_string())
                raise Exception("Case not yet handeld")
            x = self.solution.get_values()
            return x, min_cx, status

        except CplexError as exc:
            if not exc.args[2] == 1217:
                logging.error(exc)
            min_cx = nan
            x = [nan] * self.variables.get_num()
            return x, min_cx, ERROR

    def slim_solve(self) -> float:
        """Solve the MILP or LP, but return only the optimal value
                
        Example:
            optim = cplex.slim_solve()
        
        Returns:
            (float)
            
            Optimum value of the objective function.
        """
        try:
            super().solve()  # call parent solve function (that was overwritten in this class)
            status = self.solution.get_status()
            if status in [1, 101, 102, 107, 115, 128, 129, 130]:  # solution integer optimal (tolerance)
                opt = self.solution.get_objective_value()
            elif status in [118, 119]:  # solution unbounded (or inf or unbdd)
                opt = -inf
            elif status in [103, 108]:  # infeasible
                opt = nan
            else:
                logging.exception(status)
                logging.exception(self.solution.get_status_string())
                raise Exception("Case not yet handeld")
            return opt
        except CplexError as exc:
            return nan

    def populate(self, n) -> Tuple[List, float, float]:
        """Generate a solution pool for MILPs
                
        Example:
            sols_x, optim, status = cplex.populate()
        
        Returns:
            (Tuple[List of lists, float, float])
            
            solution_vectors, optimal_value, optimization_status
        """
        try:
            if isinf(n):
                self.parameters.mip.pool.capacity.set(self.parameters.mip.pool.capacity.max())
            else:
                self.parameters.mip.pool.capacity.set(n)
            self.populate_solution_pool()  # call parent solve function (that was overwritten in this class)
            status = self.solution.get_status()
            if status in [101, 102, 115, 128, 129, 130]:  # solution integer optimal
                min_cx = self.solution.get_objective_value()
                status = OPTIMAL
            elif status == 108:  # timeout without solution
                x = []
                min_cx = nan
                status = TIME_LIMIT
                return x, min_cx, status
            elif status == 103:  # infeasible
                x = []
                min_cx = nan
                status = INFEASIBLE
                return x, min_cx, status
            elif status == 107:  # timeout with solution
                min_cx = self.solution.get_objective_value()
                status = TIME_LIMIT_W_SOL
            elif status in [118, 119]:  # solution unbounded
                min_cx = -inf
                status = UNBOUNDED
            else:
                logging.exception(status)
                logging.exception(self.solution.get_status_string())
                raise Exception("Case not yet handeld")
            x = [self.solution.pool.get_values(i) for i in range(self.solution.pool.get_num())]
            return x, min_cx, status

        except CplexError as exc:
            if not exc.args[2] == 1217:
                logging.error(exc)
            min_cx = nan
            x = []
            return x, min_cx, ERROR

    def set_objective(self, c):
        """Set the objective function with a vector"""
        self.objective.set_linear([[i, c[i]] for i in range(len(c))])

    def set_objective_idx(self, C):
        """Set the objective function with index-value pairs
        
        e.g.: C=[[1, 1.0], [4,-0.2]]"""
        self.objective.set_linear(C)

    def set_ub(self, ub):
        """Set the upper bounds to a given vector"""
        self.variables.set_upper_bounds(ub)

    def set_time_limit(self, t):
        """Set the computation time limit (in seconds)"""
        if isinf(t):
            self.parameters.timelimit.set(self.parameters.timelimit.max())
        else:
            self.parameters.timelimit.set(t)

    def add_ineq_constraints(self, A_ineq, b_ineq):
        """Add inequality constraints to the model
        
        Additional inequality constraints have the form A_ineq * x <= b_ineq.
        The number of columns in A_ineq must match with the number of variables x
        in the problem.
        
        Args:
            A_ineq (sparse.csr_matrix):
                The coefficient matrix
                
            b_ineq (list of float):
                The right hand side vector
        """
        numconst = self.linear_constraints.get_num()
        numnewconst = A_ineq.shape[0]
        newconst_idx = [numconst + i for i in range(numnewconst)]
        for i, v in enumerate(b_ineq):
            if isinf(v):
                b_ineq[i] = infinity
        self.linear_constraints.add(rhs=b_ineq, senses='L' * numnewconst)
        # retrieve row and column indices from sparse matrix and convert them to int
        A_ineq = A_ineq.tocoo()
        rows_A = [int(a) + numconst for a in A_ineq.row]
        cols_A = [int(a) for a in A_ineq.col]
        # convert matrix coefficients to float
        data_A = [float(a) for a in A_ineq.data]
        self.linear_constraints.set_coefficients(zip(rows_A, cols_A, data_A))

    def add_eq_constraints(self, A_eq, b_eq):
        """Add equality constraints to the model
        
        Additional equality constraints have the form A_eq * x = b_eq.
        The number of columns in A_eq must match with the number of variables x
        in the problem.
        
        Args:
            A_eq (sparse.csr_matrix):
                The coefficient matrix
                
            b_eq (list of float):
                The right hand side vector
        """
        numconst = self.linear_constraints.get_num()
        numnewconst = A_eq.shape[0]
        self.linear_constraints.add(rhs=b_eq, senses='E' * numnewconst)
        # retrieve row and column indices from sparse matrix and convert them to int
        A_eq = A_eq.tocoo()
        rows_A = [int(a) + numconst for a in A_eq.row]
        cols_A = [int(a) for a in A_eq.col]
        # convert matrix coefficients to float
        data_A = [float(a) for a in A_eq.data]
        self.linear_constraints.set_coefficients(zip(rows_A, cols_A, data_A))

    def set_ineq_constraint(self, idx, a_ineq, b_ineq):
        """Replace a specific inequality constraint
        
        Replace the constraint with the index idx with the constraint a_ineq*x ~ b_ineq
        
        Args:
            idx (int):
                Index of the constraint
                
            a_ineq (list of float):
                The coefficient vector
                
            b_ineq (float):
                The right hand side value
        """
        if isinf(b_ineq):
            b_ineq = infinity
        self.linear_constraints.set_coefficients(zip([idx] * len(a_ineq), range(len(a_ineq)), a_ineq))
        self.linear_constraints.set_rhs([[idx, b_ineq]])



"""Test if all strain design functions run correctly."""
from .test_01_load_models_and_solvers import *
import straindesign as sd
from numpy import inf


@pytest.mark.timeout(15)
def test_mcs(curr_solver, model_small_example, comp_approach, bigM, compression):
    modules = [sd.SDModule(model_small_example, SUPPRESS, constraints=["R3 - 0.5 R1 <= 0.0", "R2 <= 0", "R1 >= 0.1"])]
    modules += [
        sd.SDModule(model_small_example, SUPPRESS, constraints=["1.0 R3 - 0.5 R1 - 0.5 R2 <= 0.0 ", "1.0 R2 >= 0.0 ", "1.0 R1 >= 0.1 "])
    ]
    modules += [sd.SDModule(model_small_example, PROTECT, constraints=["1.0 R3 >= 1.0 "])]
    kicost = {
        'R2': 1,
    }
    sd_setup = {
        MODULES: modules,
        MAX_COST: inf,
        MAX_SOLUTIONS: inf,
        SOLUTION_APPROACH: comp_approach,
        KICOST: kicost,
        SOLVER: curr_solver,
        'compress': compression,
        'M': bigM
    }
    solution = sd.compute_strain_designs(model_small_example, sd_setup=sd_setup)
    sols = solution.get_reaction_sd()
    assert ({'R1': -1.0, 'R2': 1.0} in sols)
    assert ({'R6': -1.0, 'R8': -1.0} in sols)
    assert ({'R4': -1.0} in sols)
    assert ({'R7': -1.0} in sols)
    assert ({'R10': -1.0} in sols)


@pytest.mark.timeout(15)
def test_mcs_opt(curr_solver, model_weak_coupling, comp_approach, bigM, compression):
    """Test MCS computation with nested optimization constraints."""
    modules = [sd.SDModule(model_weak_coupling, SUPPRESS, inner_objective="r_BM", constraints=["r_P - 0.4 r_S <= 0", "r_S >= 0.1"])]
    modules += [sd.SDModule(model_weak_coupling, PROTECT, constraints=["r_BM >= 0.2"])]
    kocost = {'r1': 1, 'r2': 1, 'r4': 1.1, 'r5': 0.75, 'r7': 0.8, 'r8': 1, 'r9': 1, 'r_S': 1.0, 'r_P': 1, 'r_BM': 1, 'r_Q': 1.5}
    kicost = {
        'r3': 0.6,
        'r6': 1.0,
    }
    regcost = {'r6 >= 4.5': 1.2}
    sd_setup = {
        MODULES: modules,
        MAX_COST: 4,
        MAX_SOLUTIONS: inf,
        SOLUTION_APPROACH: comp_approach,
        KOCOST: kocost,
        KICOST: kicost,
        REGCOST: regcost,
        SOLVER: curr_solver,
        'compress': compression,
        'M': bigM
    }
    solution = sd.compute_strain_designs(model_weak_coupling, sd_setup=sd_setup)
    assert (len(solution.reaction_sd) == 3)


@pytest.mark.timeout(15)
def test_mcs_gpr(model_gpr, comp_approach):
    """Test MCS computation with gpr rules."""
    modules = [sd.SDModule(model_gpr, SUPPRESS, constraints=["1.0 rd_ex >= 1.0 "])]
    modules += [sd.SDModule(model_gpr, PROTECT, constraints=[[{'r_bm': 1.0}, '>=', 1.0]])]
    kocost = {'rs_up': 1.0, 'rd_ex': 1.0, 'rp_ex': 1.1, 'r_bm': 0.75}
    gkocost = {
        'g1': 1.0,
        'g2': 1.0,
        'g4': 3.0,
        'g5': 2.0,
        'g6': 1.0,
        'g7': 1.0,
        'g8': 1.0,
        'g9': 1.0,
    }
    gkicost = {
        'g3': 1.0,
    }
    regcost = {'g4 <= 0.4': 1.2}
    sd_setup = {
        MODULES: modules,
        MAX_COST: 2,
        MAX_SOLUTIONS: inf,
        SOLUTION_APPROACH: comp_approach,
        KOCOST: kocost,
        GKOCOST: gkocost,
        GKICOST: gkicost,
        REGCOST: regcost,
        SOLVER: 'gurobi'
    }
    solution = sd.compute_strain_designs(model_gpr, sd_setup=sd_setup)
    assert (len(solution.gene_sd) == 4)


@pytest.mark.timeout(15)
def test_mcs_gpr2(model_gpr, comp_approach):
    """Test MCS computation gpr rule (compression)."""
    modules = [sd.SDModule(model_gpr, SUPPRESS, constraints=["1.0 rd_ex >= 1.0 "])]
    modules += [sd.SDModule(model_gpr, PROTECT, constraints=[[{'r_bm': 1.0}, '>=', 1.0]])]
    kocost = {'rs_up': 1.0, 'rd_ex': 1.0, 'rp_ex': 1.1, 'r_bm': 0.75}
    gkocost = {
        'G_g1': 1.0,
        'G_g2': 1.0,
        'G_g4': 3.0,
        'G_g5': 2.0,
        'G_g6': 1.0,
        'G_g7': 1.0,
        'G_g8': 1.0,
        'G_g9': 1.0,
    }
    gkicost = {
        'G_g3': 1.0,
    }
    regcost = {'G_g4 <= 0.4': 1.2}
    sd_setup = {
        MODULES: modules,
        MAX_COST: 2,
        MAX_SOLUTIONS: inf,
        SOLUTION_APPROACH: comp_approach,
        KOCOST: kocost,
        GKOCOST: gkocost,
        GKICOST: gkicost,
        REGCOST: regcost,
        SOLVER: 'cplex'
    }
    solution = sd.compute_strain_designs(model_gpr, sd_setup=sd_setup)
    assert (len(solution.gene_sd) == 4)
    assert (any(['G_g4 <= 0.4' in sol for sol in solution.get_gene_sd()]))


@pytest.mark.timeout(15)
def test_optknock(curr_solver, model_weak_coupling, comp_approach_best_populate, bigM, compression):
    """Test OptKnock computation."""
    modules = [sd.SDModule(model_weak_coupling, OPTKNOCK, outer_objective='r_P', inner_objective='r_BM',
                           constraints='r_BM>=1')]  # constraints=[[{'r_BM':1.0}, '>=' ,1.0 ]]
    kocost = {'r1': 1, 'r2': 1, 'r4': 1.1, 'r5': 0.75, 'r7': 0.8, 'r8': 1, 'r9': 1, 'r_S': 1.0, 'r_P': 1, 'r_BM': 1, 'r_Q': 1.5}
    kicost = {
        'r3': 0.6,
        'r6': 1.0,
    }
    regcost = {'r6 >= 4.5': 1.2}
    sd_setup = {
        MODULES: modules,
        MAX_COST: 4,
        MAX_SOLUTIONS: 3,
        SOLUTION_APPROACH: comp_approach_best_populate,
        KOCOST: kocost,
        KICOST: kicost,
        REGCOST: regcost,
        SOLVER: curr_solver,
        'compress': compression,
        'M': bigM
    }
    solution = sd.compute_strain_designs(model_weak_coupling, sd_setup=sd_setup)
    assert (len(solution.reaction_sd) == 3)


@pytest.mark.timeout(15)
def test_robustknock(curr_solver, model_weak_coupling, comp_approach_best_populate, bigM, compression):
    """Test RobustKnock computation."""
    modules = [
        sd.SDModule(model_weak_coupling,
                    ROBUSTKNOCK,
                    outer_objective='r_P',
                    inner_objective='r_BM',
                    constraints=[[{
                        'r_BM': 1.0
                    }, '>=', 1.0]])
    ]
    kocost = {'r1': 1, 'r2': 1, 'r4': 1.1, 'r5': 0.75, 'r7': 0.8, 'r8': 1, 'r9': 1, 'r_S': 1.0, 'r_P': 1, 'r_BM': 1, 'r_Q': 1.5}
    kicost = {
        'r3': 0.6,
        'r6': 1.0,
    }
    regcost = {'r6 >= 4.5': 1.2}
    sd_setup = {
        MODULES: modules,
        MAX_COST: 4,
        MAX_SOLUTIONS: 2,
        SOLUTION_APPROACH: comp_approach_best_populate,
        KOCOST: kocost,
        KICOST: kicost,
        REGCOST: regcost,
        SOLVER: curr_solver,
        'compress': compression,
        'M': bigM
    }
    solution = sd.compute_strain_designs(model_weak_coupling, sd_setup=sd_setup)
    for s in solution.get_reaction_sd():
        m = model_weak_coupling.copy()
        for itv, v in s.items():
            if v == -1.0:
                m.reactions.get_by_id(itv).bounds = (0, 0)
        sol_max_BM = sd.fba(m, obj='r_BM')
        sol_min_P = sd.fba(m, obj_sense='minimize', obj='r_P', constraints="r_BM >= " + str(sol_max_BM.objective_value))
        assert (sol_min_P.objective_value > 0)


@pytest.mark.timeout(15)
def test_optcouple(curr_solver, model_weak_coupling, comp_approach_best_populate, bigM, compression):
    """Test OptCouple computation."""
    modules = [sd.SDModule(model_weak_coupling, OPTCOUPLE, prod_id='r_P', inner_objective='r_BM', min_gcp=1.0)]
    kocost = {'r1': 1, 'r2': 1, 'r4': 1.1, 'r5': 0.75, 'r7': 0.8, 'r8': 1, 'r9': 1, 'r_S': 1.0, 'r_P': 1, 'r_BM': 1, 'r_Q': 1.5}
    kicost = {
        'r3': 0.6,
        'r6': 1.0,
    }
    regcost = {'r6 >= 4.5': 1.2}
    sd_setup = {
        MODULES: modules,
        MAX_COST: 6,
        MAX_SOLUTIONS: 3,
        SOLUTION_APPROACH: comp_approach_best_populate,
        KOCOST: kocost,
        KICOST: kicost,
        REGCOST: regcost,
        SOLVER: curr_solver,
        'compress': compression,
        'M': bigM
    }
    solution = sd.compute_strain_designs(model_weak_coupling, sd_setup=sd_setup)
    assert (len(solution.get_reaction_sd()) == 2)
    for s in solution.get_reaction_sd():
        constraints = []
        for itv, v in s.items():
            if v == -1.0:
                constraints += [[{itv: 1}, '=', 0]]
        constraints_no_P = constraints + [[modules[0][PROD_ID], '=', 0]]
        sol_max_BM_no_P = sd.fba(model_weak_coupling, obj='r_BM', constraints=constraints_no_P)
        max_BM_no_P = sol_max_BM_no_P.objective_value
        sol_max_BM = sd.fba(model_weak_coupling, obj='r_BM', constraints=constraints)
        max_BM = sol_max_BM.objective_value
        assert (max_BM - max_BM_no_P >= modules[0][MIN_GCP])
        constraints_max_BM = constraints + [[{'r_BM': 1}, '=', max_BM]]
        sol_min_P = sd.fba(model_weak_coupling, obj_sense='minimize', obj='r_P', constraints=constraints_max_BM)
        assert (sol_min_P.objective_value > 0)
    pass



import pytest
from cobra import Configuration
from straindesign.names import *

cobra_conf = Configuration()
bound_thres = max((abs(cobra_conf.lower_bound), abs(cobra_conf.upper_bound)))

# Initialize an empty list for solvers
solvers = [GLPK]

# Add GRUOBI to the list if the cplex package is installed
try:
    import gurobipy
    solvers.append(GUROBI)
except ImportError:
    pass  # GUROBI is not installed

# Add CPLEX to the list if the cplex package is installed
try:
    import cplex
    solvers.append(CPLEX)
except ImportError:
    pass  # CPLEX is not installed

# Add SCIP to the list if the pyscipopt package is installed
try:
    import pyscipopt
    solvers.append(SCIP)
except ImportError:
    pass  # SCIP is not installed


@pytest.fixture(params=solvers, scope="session")
def curr_solver(request: pytest.FixtureRequest) -> str:
    """Provide session-level fixture for parametrized solver names."""
    return request.param


@pytest.fixture(params=[ANY, BEST, POPULATE], scope="session")
def comp_approach(request: pytest.FixtureRequest) -> str:
    """Provide session-level fixture for computation modes."""
    return request.param


@pytest.fixture(params=[BEST, POPULATE], scope="session")
def comp_approach_best_populate(request: pytest.FixtureRequest) -> str:
    """Provide session-level fixture for nested-opt computation modes."""
    return request.param


@pytest.fixture(params=[None, bound_thres], scope="session")
def bigM(request: pytest.FixtureRequest) -> str:
    """Provide session-level fixture for nested-opt computation modes."""
    return request.param


@pytest.fixture(scope="session")
def compression(request: pytest.FixtureRequest) -> str:
    """Provide session-level fixture for nested-opt computation modes."""
    return False



#!/usr/bin/env python3
#
# Copyright 2022 Max Planck Insitute Magdeburg
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
#
#
"""Class: strain design module (SDModule)"""

from numpy import all
from copy import deepcopy
from typing import List, Dict, Tuple, Union, Set, FrozenSet
from straindesign.parse_constr import *
from straindesign.names import *


class SDModule(Dict):
    """
    Strain design modules are used to specify the goal of a strain design computation
    
    (Lists of) SDModule objects are passed to compute_strain_design to specify the goal strain design computation.
    Strain design modules indicate the appraoch that should be used (OptKnock, RobustKnock, OptCouple or MCS) and
    the parameters for each approach. In each strain design computation, one of the following modules can be used
    at most once: OptKnock, RobustKnock and OptCouple. Additionally an arbitrary number of MCS modules (PROTECT or 
    SUPPRESS) may me used. The global objective of a strain design computation depends on the specified modules.
    If an OptKnock or RobustKnock module is used, the global objective function will be defined by 'outer_objective'
    and 'outer_opt_sense'. If OptCouple is used, the global objective is derived from by 'inner_objective' and 
    'inner_opt_sense'. If only MCS-like modules (suppress, protect) are used in a computation, the number of
    interventions is globally minimized.
    
    In the following, the modules and their mandatory/optional arguments are presented in detail.
    
    OptKnock:
        Globally maximize an *outer objective* subjected to the maximization/optimization of an *inner objective*. 
        For instance, maximize product synthesis, assuming that the strain maximizes its growth rate. When used 
        with the 'best' or 'populate' approach (see compute_strain_designs), this module will guarantee that the
        found intervention set allows for the highest possible outer objective (e.g., production) under the premise
        that the inner objective (growth) is forced to be maximal. This means that the *production potential* is
        maximal. However it does not necessarily mean that production is enforced at high growth rates. For enforced
        growth coupling, refer to the other module types. Additional constraints can be used to impose certain
        properties on the designed strains: constraints='growth >= 0.5' will guarantee that the designed strain
        can reach growth rates above 0.5. constraints=['growth >= 0.5', 'EX_byprod_e <= 2'] additionally guarantees
        that the synthesis rate of a by-product stays below 2 at growth maximal flux states. Alternative inner or
        outer objective functions (e.g., ATP maintenance) can be used for diffent types of strain design.
                
        mandatory arguments: model, module_type='optknock', inner_objective, outer_objective
        optional arguments: constraints, inner_opt_sense, outer_opt_sense, skip_checks
        (Detailed description of the arguments follow below)
    
    RobustKnock:
        Globally minimize the maximum of an *outer objective* subjected to the maximization/optimization of an *inner 
        objective*. For instance, maximize the minimal product synthesis rate, assuming that the strain maximizes its 
        growth rate. When used with the 'best' or 'populate' approach (see compute_strain_designs), this module will 
        guarantee that the found intervention set enforces that the minimum of the outer objective (e.g., maximal 
        production) is maximal (max-min), given that the inner objective (growth) is forced to be maximal. This means
        that the *minimal guaranteed production* is maximal and production is guaranteed at growth-maximal flux states.
        Additional constraints can be used to impose certain properties on the designed strains: 
        constraints='growth >= 0.5' will guarantee that the designed strain can reach growth rates above 0.5/h. 
        constraints=['growth >= 0.5', 'EX_byprod_e<=2'] additionally guarantees that the synthesis rate of a by-product 
        stays below 2 mmol/gCDW/h at growth maximal flux states.

        mandatory arguments: model, module_type='robustknock', inner_objective, outer_objective
        optional arguments: constraints, inner_opt_sense, outer_opt_sense, skip_checks, reac_ids
        (Detailed description of the arguments follow below)
        
    OptCouple:
        Globally maximize the *growth-coupling potential*, that is, the difference between the maximal growth rate 
        without product synthesis and the maximum overall growth rate (with product synthesis). This strain design
        approach often leads to directionally growth-coupled strain designs. Again, alternative definitions of this 
        objective are possible to, for instance, couple production to the synthesis of ATP. To specify the product
        for which coupling should be engineered, the reaction identifier of the product exchange (pseudo)reaction
        is passed through the *prod_id* parameter (see below). One may addidionally define a minimum growth-coupling
        potential through the parameter *min_gcp*.
                
        mandatory arguments: model, module_type='optcouple', inner_objective, prod_id
        optional arguments: constraints, inner_opt_sense, min_gcp, skip_checks, reac_ids
        (Detailed description of the arguments follow below)
        
    Suppress:
        MCS-like suppression of a subspace of all steady-state flux vectors. The 'constraints' parameter is used to
        descirbe the flux states that should not be eliminated from the flux space. Depending on the goal of the
        strain design computation, undesired flux states may be those with production of an undesired by-product,
        low product synthesis rates, low product yields or even flux states with microbial growth. In addition to
        constraints, an inner objective function can be defined that is enforced. In that case flux states are
        suppressed that are optimal with respect to the specified inner objective function and additionally fulfil
        the specified constraints. If the goal is to find minimal sets of knockouts that are lethal for an organism,
        one can use a suppress module with the constraint parameter: constraints='growth >= 0.01'. The algorithm
        will then find thes smallest sets of knockouts that render flux states with growth >= 0.01 infeasible. If we
        take the example of production strain design, one could use the suppress module to enforce a certain product 
        yield (prod/subst > min_yield) growth-maximal flux states. This can be done by defining an inner objective 
        function for optmizing growth and selecting a minimum production threshold to be attained at maximum growth: 
        inner_objective='1.0 growth', constraints='EX_prod_e - min_yield*UP_subst_e  <= 0'. If used without any
        constraints, the suppress module ensures that the model is infeasible.
        
        mandatory arguments: model, module_type='suppress'
        optional arguments: constraints, inner_objective, inner_opt_sense, skip_checks, reac_ids
        (Detailed description of the arguments follow below)
        
    Protect:
        MCS-like protection of a subspace of all steady-state flux vectors. The 'constraints' parameter is used to
        descirbe flux states that should become/stay feasible when/by introducing interventions. This can be used
        to maintain or protect certain metabolic functions, like microbial growth, despite engineering a strain for 
        bioroduction. When provided with an inner objective, the protect module will ensure that flux vectors
        optimal with respect to the inner objective function will be able to fulfil the constraints set in the
        'constraints' parameter. This can be used to design potentially growth-coupled strains with a minimum set
        of metabolic interventions. As an example, if one uses the protect module to ensure that growth with 
        rates above 0.1/h remains feasible, one sets constraints='growth >= -0.1'. If the goal is to ensure that
        product synthesis is possible at rates of more than 5 mmol/gCDW/h at maximal growth, one sets:
        inner_objective='1.0 growth' and constraints='EX_prod_e >= 5'. If used without any constraints, the protect
        module just ensures that the model is feasible.
        
        mandatory arguments: model, module_type='protect'
        optional arguments: constraints,inner_objective, inner_opt_sense, skip_checks, reac_ids
        (Detailed description of the arguments follow below)

    Example:
        m = SDModule(model,'optknock',outer_objective='growth', inner_objective='EX_etoh_e', constraints='growth >= 0.2')

    Args:
        model (cobra.Model):
        
            A metabolic model that is an instance of the cobra.Model class. Instead of a model, a dummy
            object can be used as long as it has the field 'id'. If a dummy is used, skip_checks=True
            must be used and a list of reaction ids must be provided through reac_ids=*list_of_strings*.
            
        module_type (str):
        
            A string that specifies the module type. Allowed values are 'optknock', 'robustknock', 
            'optcouple', 'protect', 'suppress'. Depending on the specified module type, other parameters
            must be set accordingly (see description above).
            
        constraints (optional (str) or (list of str) or (list of [dict,str,float])): (Default: '')
        
            List of *linear* constraints to be used in the module, e.g., to be enforced, suppressed or taken 
            into account: signs + or -, scalar factors for reaction rates, inclusive (in)equalities and a
            float value on the right hand side. The parsing of the constraints input allows for some 
            flexibility. Correct (and identical) inputs are, for instance: 
            constraints='-EX_o2_e <= 5, ATPM = 20' or
            constraints=['-EX_o2_e <= 5', 'ATPM = 20'] or
            constraints=[[{'EX_o2_e':-1},'<=',5], [{'ATPM':1},'=',20]]
            
        inner_objective (optional (str) or (dict)):

            The *linear* inner objective function for any module type. This parameter is mandatory for 
            OptKnock, RobustKnock and OptCouple modules and optional for suppress and protect modules. If 
            used, an optimization is nested into the global problem. It can be used to account for the 
            biological objective of growth (inner_objective='BIOMASS_Ecoli_core_w_GAM'). Any linear 
            expression can be used as input, either as a single string or as a dict. Correct (and identical) 
            inputs are, for instance:
            inner_objective='BIOMASS_Ecoli_core_w_GAM - 0.05 EX_etoh_e'
            inner_objective={'BIOMASS_Ecoli_core_w_GAM': 1, 'EX_etoh_e': -0.05}
            
        inner_opt_sense (optional (str)): (Default: 'maximize')
        
            Sense of the inner optimization (maximization or minimization). Allowed values are 'minimize'
            and 'maximize'.

        outer_objective (optional (str) or (dict)):

            The *linear* outer objective function for any module type. This parameter is mandatory for 
            OptKnock and RobustKnock and cannot be used with OptCouple, suppress and protect modules. If 
            applied, this objective function is used as the global objective function. In case of OptCouple,
            the global objective is the maximization of the growth-coupling potential and the outer objective
            is not specified manually. Typical outer objectives are the optimization of product synthesis
            (outer_objective='BIOMASS_Ecoli_core_w_GAM'), but also combinations of product synthesis and 
            growth are possible. Any linear expression can be used as input, either as a single string or as 
            a dict. Correct (and identical) inputs are, for instance:
            outer_objective='BIOMASS_Ecoli_core_w_GAM + 0.1 EX_prod_e'
            outer_objective={'BIOMASS_Ecoli_core_w_GAM': 1, 'EX_prod_e': 0.1}
            
        outer_opt_sense (optional (str)): (Default: 'maximize')
        
            Sense of the outer optimization (maximization or minimization). Allowed values are 'minimize'
            and 'maximize'.
            
        prod_id (optional (str) or (dict)):
        
            The reaction id of the product of interest. This parameter is *only* used in OptCouple strain
            design and will have no effect as part of any other module. Permitted is any linear expression
            either in the form of a string or as a dict:
            prod_id='EX_etoh'
            prod_id={'EX_etoh': 1}
            
        min_gcp (optional (float)): (Default: 0.0)
        
            Minimial growth-coupling potential (GCP). I.e., the minimum difference between maximum growth
            with and without production. In practice there are two nested optimizations, one that optimizes
            inner_objective and one that optimizes inner_objective and additionally demands that the
            constraint: prod_id=0 holds. Therefore, min_gcp presents the minimum enforced growth-coupling
            potential, so, a minimum objective value. This parameter is supposed to be used with the 'any'
            approach and has (virtually) no effect when used with 'best' or 'populate', since GCP is
            maximized anyway.
            
        skip_checks (optional (bool)): (Default: False)

            Skip the module verification. If checks are not skipped, the constructor will verify, if the
            module is feasible with the original model, that is, if all entered parameters parse correctly
            and if the given sets of constraints can be applied on the model without rendering it infeasible.
            Finally, it throws an error if the trivial 0-vector is feasible in the subspaces defined by a
            suppress or protect module, since the user should ensure that the 0-vector is excluded from these
            subspaces (see online documentation for detail).

    Returns:
        (SDModule):
            A strain design module object that can be used with the function compute_strain_design.
            Multiple modules can be used to specify a strain design problem.
    """

    def __init__(self, model, module_type, *args, **kwargs):
        self[MODEL_ID] = model.id
        self[MODULE_TYPE] = module_type
        allowed_keys = {
            CONSTRAINTS, INNER_OBJECTIVE, INNER_OPT_SENSE, OUTER_OBJECTIVE, OUTER_OPT_SENSE, PROD_ID, 'skip_checks', MIN_GCP, 'reac_ids'
        }
        # set all keys passed in kwargs as properties of the SD_Module object
        for key, value in kwargs.items():
            if key in allowed_keys:
                self[key] = value
            else:
                raise Exception("Key " + key + " is not supported.")
        # set all undefined keys to None
        for key in allowed_keys:
            if key not in kwargs.keys():
                self[key] = None

        if not self['reac_ids'] and not model.reactions:
            raise Exception('Strain design module cannot be constructed without information about '+\
                            'available reactions reactions. Make sure to provide a valid model or '+\
                            'reaction list.')

        # check if there is sufficient information for each module type
        if self[MODULE_TYPE] not in [PROTECT, SUPPRESS, OPTKNOCK, ROBUSTKNOCK, OPTCOUPLE]:
            raise Exception('"' + MODULE_TYPE + '" must be "' + PROTECT + '", "' + SUPPRESS + '", "' + OPTKNOCK + '", "' + ROBUSTKNOCK +
                            '", "' + OPTCOUPLE + '".')
        if (self[MODULE_TYPE] in [OPTKNOCK, ROBUSTKNOCK]):
            if self[INNER_OPT_SENSE] is None:
                self[INNER_OPT_SENSE] = MAXIMIZE
            if self[OUTER_OPT_SENSE] is None:
                self[OUTER_OPT_SENSE] = MAXIMIZE
            elif self[INNER_OPT_SENSE] not in [MINIMIZE, MAXIMIZE] or self[OUTER_OPT_SENSE] not in [MINIMIZE, MAXIMIZE]:
                raise Exception('Inner and outer optimization sense must be "' + MINIMIZE + '" or "' + MAXIMIZE + '" (default).')
            if ((self[INNER_OBJECTIVE] == None) or (self[OUTER_OBJECTIVE] == None)):
                raise Exception('When module type is "' + OPTKNOCK + '" or "' + ROBUSTKNOCK +
                                '", an inner and outer objective function must be provided.')
        elif (self[MODULE_TYPE] == OPTCOUPLE):
            if self[INNER_OPT_SENSE] is None:
                self[INNER_OPT_SENSE] = MAXIMIZE
            if self[INNER_OPT_SENSE] not in [MINIMIZE, MAXIMIZE]:
                raise Exception('Inner optimization sense must be "' + MINIMIZE + '" or "' + MAXIMIZE + '" (default).')
            if self[MIN_GCP] is None:
                self[MIN_GCP] = 0.0
            if self[INNER_OBJECTIVE] == None:
                raise Exception('When module type is "' + OPTCOUPLE + '", an inner objective function must be provided.')
            if self[PROD_ID] == None:
                raise Exception('When module type is "' + OPTCOUPLE + '", the production reaction id must be provided.')

        if not self['reac_ids']:
            self['reac_ids'] = model.reactions.list_attr('id')

        # parse constraints and ensure they have the form:
        # [ [{'r1': -1, 'r3': 2}, '<=', 3],
        #   [{'r2': 1, 'r3': -1},  '=', 0]  ]
        if self[CONSTRAINTS] is not None:
            self[CONSTRAINTS] = parse_constraints(self[CONSTRAINTS], self['reac_ids'])
        else:
            self[CONSTRAINTS] = []

        # parse inner objective
        if self[INNER_OBJECTIVE] is not None:
            if type(self[INNER_OBJECTIVE]) is str:
                self[INNER_OBJECTIVE] = linexpr2dict(self[INNER_OBJECTIVE], self['reac_ids'])

        # parse outer objective
        if self[OUTER_OBJECTIVE] is not None:
            if type(self[OUTER_OBJECTIVE]) is str:
                self[OUTER_OBJECTIVE] = linexpr2dict(self[OUTER_OBJECTIVE], self['reac_ids'])

        # parse prod_id
        if self[PROD_ID] is not None:
            if type(self[PROD_ID]) is str:
                self[PROD_ID] = linexpr2dict(self[PROD_ID], self['reac_ids'])

        # verify self[CONSTRAINTS]
        if not self['skip_checks']:
            from straindesign import fba
            if fba(model, constraints=self[CONSTRAINTS]).status == INFEASIBLE:
                raise Exception("There is no feasible solution of the model under the given constraints.")

            if self[MODULE_TYPE] in [SUPPRESS, PROTECT] and self[INNER_OBJECTIVE] is not None:
                constr = [[{k: 1}, '=', 0] for k in model.reactions.list_attr('id')]
                if fba(model, constraints=self[CONSTRAINTS] + constr).status != INFEASIBLE:
                    raise Exception('When '+MODULE_TYPE+' is "'+SUPPRESS+\
                        '", the zero vector must not be a contained in the described flux space.')

            if (self[INNER_OBJECTIVE]
                    is not None) and (not all([True if r in self['reac_ids'] else False for r in self[INNER_OBJECTIVE].keys()])):
                raise Exception("Inner objective invalid.")

            if (self[OUTER_OBJECTIVE]
                    is not None) and (not all([True if r in self['reac_ids'] else False for r in self[OUTER_OBJECTIVE].keys()])):
                raise Exception("Outer objective invalid.")

            if (self[PROD_ID] is not None) and (not all([True if r in self['reac_ids'] else False for r in self[PROD_ID].keys()])):
                raise Exception("Production id (prod_id) invalid.")

            if (self[MIN_GCP] is not None) and type(self[MIN_GCP]) is not None:
                if type(self[MIN_GCP]) is float:
                    pass
                elif type(self[MIN_GCP]) is int:
                    self[MIN_GCP] = float(self[MIN_GCP])
                else:
                    raise Exception("Minimum growth coupling potential (" + MIN_GCP + ").")

    def copy(self):
        """Create a deep copy of a strain design module."""

        class DummyModel:
            id = self[MODEL_ID]

        return SDModule(DummyModel(),
                        self[MODULE_TYPE],
                        constraints=deepcopy(self[CONSTRAINTS]),
                        inner_objective=deepcopy(self[INNER_OBJECTIVE]),
                        inner_opt_sense=deepcopy(self[INNER_OPT_SENSE]),
                        outer_objective=deepcopy(self[OUTER_OBJECTIVE]),
                        outer_opt_sense=deepcopy(self[OUTER_OPT_SENSE]),
                        prod_id=deepcopy(self[PROD_ID]),
                        min_gcp=self[MIN_GCP],
                        skip_checks=True,
                        reac_ids=deepcopy(self['reac_ids']))



#!/usr/bin/env python3
#
# Copyright 2022 Max Planck Insitute Magdeburg
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
#
#
"""Classes and functions for the construction of strain design MILPs

This module contains functions that help construct mixed-integer linear problems,
mainly functions that facilitate the construction of LP and Farkas dual
problems from linear problems of the type A_ineq*x<=b_ineq, A_eq*x<=b_eq, lb<=x<=ub.
The functions also help keeping track of the relationship of constraints and variables
and their individual counterparts in dual problems, which is essential when simulating 
knockouts in dual problems. Most of the time, the sparse datatype is used to store and
edit matrices for improved speed and memory."""

import numpy as np
from scipy import sparse
from cobra.util import create_stoichiometric_matrix
from cobra import Model, Configuration
from typing import List, Tuple
from straindesign import SDModule, IndicatorConstraints, lineqlist2mat, linexprdict2mat, MILP_LP, SDPool, \
                         avail_solvers, select_solver, remove_dummy_bounds, SDModule
from straindesign.names import *
import logging


class SDProblem:
    """Strain design MILP
    
    The construcor of this class translates a model and strain design modules into a mixed integer linear
    problem. This class, however, is the backbone of the strain design computation. Preprocessing steps 
    that enable gene, reaction and regulatory interventions, or network compression usually preceed the 
    construction of an SDProblem-object and are integrated in the function *compute_strain_designs*. 

    Args:
        model (cobra.Model):
            A metabolic model that is an instance of the cobra.Model class.
        
        sd_modules ((list of) straindesign.SDModule):
            Modules that specify the strain design problem, e.g., protected or suppressed flux states for 
            MCS strain design or inner and outer objective functions for OptKnock. See description
            of SDModule for more information on how to set up modules.
            
        ko_cost (optional (dict)): (Default: None)
            A dictionary of reaction identifiers and their associated knockout costs. If not specified, all reactions
            are treated as knockout candidates, equivalent to ko_cost = {'r1':1, 'r2':1, ...}. If a subset of reactions
            is listed in the dict, all other are not considered as knockout candidates.
            
        ki_cost (optional (dict)): (Default: None)
            A dictionary of reaction identifiers and their associated costs for addition. If not specified, all reactions
            are treated as knockout candidates. Reaction addition candidates must be present in the original model with
            the intended flux boundaries **after** insertion. Additions are treated adversely to knockouts, meaning that
            their exclusion from the network is not associated with any cost while their presence entails intervention costs.
            
        max_cost (optional (int)): (Default: inf): 
            The maximum cost threshold for interventions. Every possible intervention is associated with a
            cost value (1, by default). Strain designs cannot exceed the max_cost threshold. Individual
            intervention cost factors may be defined through ki_cost and ko_cost.
            
        solver (optional (str)): (Default: same as defined in model / COBRApy)
            The solver that should be used for preparing and carrying out the strain design computation.
            Allowed values are 'cplex', 'gurobi', 'scip' and 'glpk'.
            
        M (optional (int)): (Default: None)
            If this value is specified (and non-zero, not None), the computation uses the big-M 
            method instead of indicator constraints. Since GLPK does not support indicator constraints it uses
            the big-M method by default (with COBRA standard M=1000). M should be chosen 'sufficiently large' 
            to avoid computational artifacts and 'sufficiently small' to avoid numerical issues.
            
        essential_kis (optional (set)):
            A set of reactions that are marked as addable and that are essential for at least one of the
            strain design modules. Providing such "essential knock-ins" may speed up the strain design computation.
            
    Returns:
        (SDProblem):
            An instance of SDProblem containing the strain design MILP
    """

    def __init__(self, model: Model, sd_modules: List[SDModule], *args, **kwargs):
        allowed_keys = {KOCOST, KICOST, SOLVER, MAX_COST, 'M', 'essential_kis', SEED}
        # set all keys passed in kwargs
        for key, value in dict(kwargs).items():
            if key in allowed_keys:
                setattr(self, key, value)
            else:
                raise Exception("Key " + key + " is not supported.")
        # set all remaining keys to None
        for key in allowed_keys:
            if key not in dict(kwargs).keys():
                setattr(self, key, None)

        if "SD_Module" in str(type(sd_modules)):
            sd_modules = [sd_modules]

        if self.solver is None:
            if len(avail_solvers) > 0:
                self.solver = list(avail_solvers)[0]
            else:
                raise Exception('No solver available. Please ensure that one of the following '\
                    'solvers is avaialable in your Python environment: CPLEX, Gurobi, SCIP, GLPK')
        self.solver = select_solver(self.solver, model)
        cobra_conf = Configuration()
        bound_thres = max((abs(cobra_conf.lower_bound), abs(cobra_conf.upper_bound)))
        if self.M is None and self.solver == 'glpk':
            logging.warning(
                'GLPK only supports strain design computation with the bigM method. Using cobra bound: '+str(bound_thres)+\
                ' as M.')
            self.M = bound_thres
        elif self.M is None:
            self.M = np.inf
        # the matrices in sd_modules, ko_cost and ki_cost should be numpy.array or scipy.sparse (csr, csc, lil) format
        self.model = model
        self.sd_modules = sd_modules
        reac_ids = model.reactions.list_attr("id")
        numr = len(model.reactions)
        # Create vectors for ko_cost, ki_cost, inverted bool-vars and non-targetable bools
        # If no knockable reactions are assigned, assume all are KO-able.
        # Generally, KIs overwrite KOs
        if self.ko_cost is None:
            self.ko_cost = {rid: 1.0 for rid in reac_ids}
        if self.ki_cost is None:
            self.ki_cost = {}
        if self.essential_kis is None:
            self.essential_kis = set()
        self.ko_cost = [float(self.ko_cost.get(key)) if (key in self.ko_cost.keys()) else np.nan for key in reac_ids]
        self.ki_cost = [float(self.ki_cost.get(key)) if (key in self.ki_cost.keys()) else np.nan for key in reac_ids]
        self.ko_cost = [self.ko_cost[i] if np.isnan(self.ki_cost[i]) else np.nan for i in range(numr)]
        self.num_z = numr
        self.cost = [i for i in self.ko_cost]
        for i in [i for i, x in enumerate(self.ki_cost) if not np.isnan(x)]:
            self.cost[i] = self.ki_cost[i]
        self.z_inverted = [not np.isnan(x) for x in self.ki_cost]
        self.z_non_targetable = [np.isnan(x) for x in self.cost]
        for i in [i for i, x in enumerate(self.cost) if np.isnan(x)]:
            self.cost[i] = 0.0
        # Prepare top 3 lines of MILP (sum of weighted interventions below (0) and above (1) threshold) and objective function (2)
        self.idx_row_maxcost = 0
        self.idx_row_mincost = 1
        self.idx_row_obj = 2
        self.A_ineq = sparse.csr_matrix([[-i for i in self.cost], self.cost, [0 for _ in range(self.num_z)]])
        if self.max_cost is None:
            self.b_ineq = [0.0, float(np.sum(np.abs(self.cost))), np.inf]
        else:
            self.b_ineq = [0.0, float(self.max_cost), np.inf]
        self.z_map_constr_ineq = sparse.csc_matrix((numr, 3))
        self.lb = [1.0 if r in self.essential_kis else 0.0 for r in model.reactions]
        self.ub = [1.0 - float(i) for i in self.z_non_targetable]
        self.idx_z = [i for i in range(0, numr)]
        self.c = [0.0] * numr
        # Initialize also empty equality matrix
        self.A_eq = sparse.csc_matrix((0, numr))
        self.b_eq = []
        self.z_map_constr_eq = sparse.csc_matrix((numr, 0))
        self.num_modules = 0
        self.indic_constr = []  # Add instances of the class 'Indicator_constraint' later
        # Initialize association between z and variables and variables
        self.z_map_vars = sparse.csc_matrix((numr, numr))
        # replace bounds with inf if above a cobra bound threshold
        remove_dummy_bounds(self.model)
        logging.info('Constructing strain design MILP for solver: ' + self.solver + '.')
        for i in range(len(sd_modules)):
            self.addModule(sd_modules[i])

        # Assign knock-ins/outs correctly by taking into account z_inverted
        # invert *(-1) rows in z_map_constr_eq, z_map_constr_ineq, z_map_vars
        # where there are "knock-ins" / additions
        # make knock-in/out matrix
        z_kos_kis = [
            1 if (not self.z_non_targetable[i]) and (not self.z_inverted[i]) else -1 if self.z_inverted[i] else 0
            for i in range(0, self.num_z)
        ]
        z_kos_kis = sparse.diags(z_kos_kis)
        self.z_map_constr_ineq = z_kos_kis * self.z_map_constr_ineq
        self.z_map_constr_eq = z_kos_kis * self.z_map_constr_eq
        self.z_map_vars = z_kos_kis * self.z_map_vars

        # Save continous part of MILP for easy strain design validation
        cont_vars = [i for i in range(0, self.A_ineq.shape[1]) if not i in self.idx_z]
        self.cont_MILP = ContMILP(self.A_ineq[:, cont_vars], self.b_ineq.copy(), self.A_eq[:, cont_vars], self.b_eq.copy(),
                                  [self.lb[i] for i in cont_vars], [self.ub[i] for i in cont_vars], [self.c[i] for i in cont_vars],
                                  self.z_map_constr_ineq.tocoo(), self.z_map_constr_eq.tocoo(), self.z_map_vars[:, cont_vars].tocoo())

        # 4. Link LP module to z-variables
        self.link_z()

        # if there are only mcs modules, minimize the knockout costs,
        # otherwise use objective function(s) from modules
        if all([mod[MODULE_TYPE] in [PROTECT, SUPPRESS] for mod in sd_modules]):
            for i in self.idx_z:
                self.c[i] = self.cost[i]
            self.is_mcs_computation = True
        else:
            self.is_mcs_computation = False
            for i in self.idx_z:
                self.c[i] = 0.0
            self.A_ineq = self.A_ineq.tolil()
            self.A_ineq[2] = sparse.lil_matrix(self.c)  # set objective
            self.A_ineq = self.A_ineq.tocsr()

        # backup objective function
        self.c_bu = [float(i) for i in self.c.copy()]

        # # for debugging
        # A = sparse.vstack(( self.A_ineq,sparse.csr_matrix([np.nan]*len(self.c)),\
        #                     self.A_eq,sparse.csr_matrix([np.nan]*len(self.c)),\
        #                     self.indic_constr.A,sparse.csr_matrix([np.nan]*len(self.c)),\
        #                     sparse.csr_matrix(self.ub),sparse.csr_matrix(self.lb)))
        # b = self.b_ineq + [np.nan] + self.b_eq + [np.nan] + self.indic_constr.b + [np.nan, np.nan, np.nan]
        # Ab = sparse.hstack((A,sparse.csr_matrix([np.nan]*len(b)).transpose(),sparse.csr_matrix(b).transpose()))
        # np.savetxt("Ab_py.tsv", Ab.todense(), delimiter='\t')
        self.vtype = 'B' * self.num_z + 'C' * (self.z_map_vars.shape[1] - self.num_z)

    def addModule(self, sd_module):
        """Generate module LP and z-linking-matrix for each module and add them to the strain design MILP
        
        Args:
            sd_module (straindesign.SDModule):
                Modules to describe strain design problems like protected or suppressed flux states for 
                MCS strain design or inner and outer objective functions for OptKnock. See description
                of SDModule for more information on how to set up modules.
        """
        self.num_modules += 1
        z_map_constr_ineq_i = []
        z_map_constr_eq_i = []
        z_map_vars_i = []
        # 1. Translate (in)equalities into matrix form
        V_ineq, v_ineq, V_eq, v_eq = lineqlist2mat(sd_module[CONSTRAINTS], self.model.reactions.list_attr('id'))
        # 2. Construct LP for module
        if sd_module[MODULE_TYPE] in [PROTECT, SUPPRESS] and sd_module[INNER_OBJECTIVE] is None:
            # Classical MCS
            A_ineq_p, b_ineq_p, A_eq_p, b_eq_p, lb_p, ub_p, c_p, z_map_constr_ineq_p, z_map_constr_eq_p, z_map_vars_p \
                = build_primal_from_cbm(self.model, V_ineq, v_ineq, V_eq, v_eq)
        elif sd_module[MODULE_TYPE] in [PROTECT, SUPPRESS, OPTKNOCK, OPTCOUPLE]:
            c_in = linexprdict2mat(sd_module[INNER_OBJECTIVE], self.model.reactions.list_attr('id'))
            # by default, assume maximization of the inner objective
            if not hasattr(sd_module,INNER_OPT_SENSE) or sd_module[INNER_OPT_SENSE] is None or \
                sd_module[INNER_OPT_SENSE] not in [MINIMIZE, MAXIMIZE] or sd_module[INNER_OPT_SENSE] == MAXIMIZE:
                c_in = -c_in
            c_in = c_in.toarray()[0].tolist()
            # 1. build primal w/ desired constraint (build_primal_from_cbm) - also store variable c
            A_ineq_v, b_ineq_v, A_eq_v, b_eq_v, lb_v, ub_v, c_v, z_map_constr_ineq_v, z_map_constr_eq_v, z_map_vars_v \
                = build_primal_from_cbm(self.model, V_ineq, v_ineq, V_eq, v_eq, c_in)
            # 2. build primal w/o desired constraint (build_primal_from_cbm) - store c_inner
            A_ineq_inner, b_ineq_inner, A_eq_inner, b_eq_inner, lb_inner, ub_inner, c_inner, z_map_constr_ineq_inner, z_map_constr_eq_inner, z_map_vars_inner \
                = build_primal_from_cbm(self.model, V_ineq=None, v_ineq=None, V_eq=None, v_eq=None, c=c_in)
            # 3. build dual from primal w/o desired constraint (build_dual w/o the farkas-option) - store c_inner_dual
            A_ineq_dual, b_ineq_dual, A_eq_dual, b_eq_dual, lb_dual, ub_dual, c_inner_dual, z_map_constr_ineq_dual, z_map_constr_eq_dual, z_map_vars_dual \
                = LP_dualize(A_ineq_inner, b_ineq_inner, A_eq_inner, b_eq_inner, lb_inner, ub_inner, c_inner, z_map_constr_ineq_inner, z_map_constr_eq_inner, z_map_vars_inner)
            # 4. connect primal w/ undesired region and dual w/o undesired region (i.e. biomass) via c = c_inner.
            A_ineq_p = sparse.block_diag((A_ineq_v, A_ineq_dual)).tocsr()
            b_ineq_p = b_ineq_v + b_ineq_dual
            A_eq_p = sparse.vstack((sparse.block_diag(
                (A_eq_v, A_eq_dual)), sparse.hstack((sparse.csr_matrix(c_v), sparse.csr_matrix(c_inner_dual))))).tocsr()
            b_eq_p = b_eq_v + b_eq_dual + [0.0]
            lb_p = lb_v + lb_dual
            ub_p = ub_v + ub_dual
            # 5. Update z-associations
            z_map_vars_p = sparse.hstack((z_map_vars_v, z_map_vars_dual))
            z_map_constr_ineq_p = sparse.hstack((z_map_constr_ineq_v, z_map_constr_ineq_dual))
            z_map_constr_eq_p = sparse.hstack((z_map_constr_eq_v, z_map_constr_eq_dual, sparse.csc_matrix((self.num_z, 1))))
        elif sd_module[MODULE_TYPE] == ROBUSTKNOCK:
            # RobustKnock has three layers, inner maximization and an outer min-max problem
            c_in = linexprdict2mat(sd_module[INNER_OBJECTIVE], self.model.reactions.list_attr('id'))
            # by default, assume maximization of the inner objective
            if not hasattr(sd_module,INNER_OPT_SENSE) or sd_module[INNER_OPT_SENSE] is None or \
                sd_module[INNER_OPT_SENSE] not in [MINIMIZE, MAXIMIZE] or sd_module[INNER_OPT_SENSE] == MAXIMIZE:
                c_in = -c_in
            c_in = c_in.toarray()[0].tolist()
            # 1. build primal of inner problem (build_primal_from_cbm) - also store variable c
            A_ineq_v, b_ineq_v, A_eq_v, b_eq_v, lb_v, ub_v, c_v, z_map_constr_ineq_v, z_map_constr_eq_v, z_map_vars_v \
                = build_primal_from_cbm(self.model, V_ineq=None, v_ineq=None, V_eq=None, v_eq=None, c=c_in)
            # 2. build primal of outer problem (base of inner outer problem) (build_primal_from_cbm) - with inner objective
            A_ineq_inner, b_ineq_inner, A_eq_inner, b_eq_inner, lb_inner, ub_inner, c_inner, z_map_constr_ineq_inner, z_map_constr_eq_inner, z_map_vars_inner \
                = build_primal_from_cbm(self.model,V_ineq=None, v_ineq=None, V_eq=None, v_eq=None, c=c_in)
            # 3. build primal of outer problem (outer outer problem) (build_primal_from_cbm) - store c_inner
            c_out = linexprdict2mat(sd_module[OUTER_OBJECTIVE], self.model.reactions.list_attr('id'))  # get outer objective
            if not hasattr(sd_module,OUTER_OPT_SENSE) or sd_module[OUTER_OPT_SENSE] is None or \
                sd_module[OUTER_OPT_SENSE] not in [MINIMIZE, MAXIMIZE] or sd_module[OUTER_OPT_SENSE] == MAXIMIZE:
                c_out = -c_out
            c_out = c_out.toarray()[0].tolist()
            A_ineq_r, b_ineq_r, A_eq_r, b_eq_r, lb_r, ub_r, c_r, z_map_constr_ineq_r, z_map_constr_eq_r, z_map_vars_r \
                = build_primal_from_cbm(self.model,V_ineq, v_ineq, V_eq, v_eq, [-c for c in c_out])
            # 4. build dual of innerst problem - store c_inner_dual
            A_ineq_dual, b_ineq_dual, A_eq_dual, b_eq_dual, lb_dual, ub_dual, c_inner_dual, z_map_constr_ineq_dual, z_map_constr_eq_dual, z_map_vars_dual \
                = LP_dualize(A_ineq_inner, b_ineq_inner, A_eq_inner, b_eq_inner, lb_inner, ub_inner, c_inner, z_map_constr_ineq_inner, z_map_constr_eq_inner, z_map_vars_inner)
            # 5. connect primal w/ undesired region and dual w/o undesired region (i.e. biomass) via c = c_inner.
            A_ineq_p = sparse.block_diag((A_ineq_v, A_ineq_dual)).tocsr()
            b_ineq_p = b_ineq_v + b_ineq_dual
            A_eq_p = sparse.vstack((sparse.block_diag(
                (A_eq_v, A_eq_dual)), sparse.hstack((sparse.csr_matrix(c_v), sparse.csr_matrix(c_inner_dual))))).tocsr()
            b_eq_p = b_eq_v + b_eq_dual + [0.0]
            lb_p = lb_v + lb_dual
            ub_p = ub_v + ub_dual
            c_out_in_p = -sparse.csr_matrix(c_out)
            c_out_in_p.resize((1, A_ineq_p.shape[1]))
            c_out_in_p = c_out_in_p.toarray()[0].tolist()
            # 6. Update z-associations
            z_map_vars_p = sparse.hstack((z_map_vars_v, z_map_vars_dual))
            z_map_constr_ineq_p = sparse.hstack((z_map_constr_ineq_v, z_map_constr_ineq_dual))
            z_map_constr_eq_p = sparse.hstack((z_map_constr_eq_v, z_map_constr_eq_dual, sparse.csc_matrix((self.num_z, 1))))
            # 7. Dualize the joint inner problem
            A_ineq_dl_mmx, b_ineq_dl_mmx, A_eq_dl_mmx, b_eq_dl_mmx, lb_dl_mmx, ub_dl_mmx, c_dl_mmx, z_map_constr_ineq_dl_mmx, z_map_constr_eq_dl_mmx, z_map_vars_dl_mmx \
                = LP_dualize(A_ineq_p, b_ineq_p, A_eq_p, b_eq_p, lb_p, ub_p, c_out_in_p, z_map_constr_ineq_p, z_map_constr_eq_p, z_map_vars_p)
            # 8. Connect outer problem to the dualized combined inner problem to construct min-max problem.
            A_ineq_q = sparse.block_diag((A_ineq_r, A_ineq_dl_mmx)).tocsr()
            b_ineq_q = b_ineq_r + b_ineq_dl_mmx
            A_eq_q = sparse.vstack((sparse.block_diag(
                (A_eq_r, A_eq_dl_mmx)), sparse.hstack((sparse.csr_matrix(c_r), sparse.csr_matrix(c_dl_mmx))))).tocsr()
            b_eq_q = b_eq_r + b_eq_dl_mmx + [0.0]
            lb_q = lb_r + lb_dl_mmx
            ub_q = ub_r + ub_dl_mmx
            # 9. Update z-associations
            z_map_vars_q = sparse.hstack((z_map_vars_r, z_map_vars_dl_mmx))
            z_map_constr_ineq_q = sparse.hstack((z_map_constr_ineq_r, z_map_constr_ineq_dl_mmx))
            z_map_constr_eq_q = sparse.hstack((z_map_constr_eq_r, z_map_constr_eq_dl_mmx, sparse.csc_matrix((self.num_z, 1))))
            # 10. reassign lb, ub where possible
            A_ineq_i, b_ineq_i, A_eq_i, b_eq_i, lb_i, ub_i, z_map_constr_ineq_i, z_map_constr_eq_i = reassign_lb_ub_from_ineq(
                A_ineq_q, b_ineq_q, A_eq_q, b_eq_q, lb_q, ub_q, z_map_constr_ineq_q, z_map_constr_eq_q, z_map_vars_q)
            z_map_vars_i = z_map_vars_q
            c_i = sparse.csr_matrix(c_out)
            c_i.resize((1, A_ineq_i.shape[1]))
            c_i = c_i.toarray()[0].tolist()
        if sd_module[MODULE_TYPE] == OPTCOUPLE:
            c_p = c_v + [0 for _ in c_inner_dual]
            # (continued from optknock)
            prod_eq = linexprdict2mat(sd_module[PROD_ID], self.model.reactions.list_attr('id'))
            # 6.  build primal system with no production - also store variable c
            A_ineq_r, b_ineq_r, A_eq_r, b_eq_r, lb_r, ub_r, c_r, z_map_constr_ineq_r, z_map_constr_eq_r, z_map_vars_r \
                = build_primal_from_cbm(self.model,V_ineq=None, v_ineq=None, V_eq=prod_eq, v_eq=[0], c=c_in)
            # 7. Dualize no-production system.
            A_ineq_r_dl, b_ineq_dl_r_dl, A_eq_dl_r_dl, b_eq_r_dl, lb_r_dl, ub_r_dl, c_r_dl, z_map_constr_ineq_r_dl, z_map_constr_eq_r_dl, z_map_vars_r_dl \
                = LP_dualize(A_ineq_r, b_ineq_r, A_eq_r, b_eq_r, lb_r, ub_r, c_r, z_map_constr_ineq_r, z_map_constr_eq_r, z_map_vars_r)
            # 8. Create no-production bi-level system.
            A_ineq_b = sparse.block_diag((A_ineq_r, A_ineq_r_dl), format='csr')
            b_ineq_b = b_ineq_r + b_ineq_dl_r_dl
            A_eq_b = sparse.vstack((sparse.block_diag(
                (A_eq_r, A_eq_dl_r_dl)), sparse.hstack((sparse.csr_matrix(c_r), sparse.csr_matrix(c_r_dl)))),
                                   format='csr')
            b_eq_b = b_eq_r + b_eq_r_dl + [0.0]
            lb_b = lb_r + lb_r_dl
            ub_b = ub_r + ub_r_dl
            c_b = c_r + [0.0] * len(c_r_dl)
            z_map_vars_b = sparse.hstack((z_map_vars_r, z_map_vars_r_dl))
            z_map_constr_ineq_b = sparse.hstack((z_map_constr_ineq_r, z_map_constr_ineq_r_dl))
            z_map_constr_eq_b = sparse.hstack((z_map_constr_eq_r, z_map_constr_eq_r_dl, sparse.csc_matrix((self.num_z, 1))))
            # 9. Connect optknock-problem to the no-production-bilevel system to construct the optcouple problem.
            A_ineq_q = sparse.block_diag((A_ineq_p, A_ineq_b), format='csr')
            b_ineq_q = b_ineq_p + b_ineq_b
            A_eq_q = sparse.block_diag((A_eq_p, A_eq_b), format='csr')
            b_eq_q = b_eq_p + b_eq_b
            lb_q = lb_p + lb_b
            ub_q = ub_p + ub_b
            z_map_vars_q = sparse.hstack((z_map_vars_p, z_map_vars_b))
            z_map_constr_ineq_q = sparse.hstack((z_map_constr_ineq_p, z_map_constr_ineq_b))
            z_map_constr_eq_q = sparse.hstack((z_map_constr_eq_p, z_map_constr_eq_b))
            # if minimum growth-coupling potential is specified, enforce it through inequality
            if MIN_GCP in sd_module:
                A_ineq_q = sparse.vstack((A_ineq_q, sparse.lil_matrix(c_p + [-c for c in c_b])), format='csr')
                b_ineq_q = b_ineq_q + [-sd_module[MIN_GCP]]
                z_map_constr_ineq_q = sparse.hstack((z_map_constr_ineq_q, sparse.csc_matrix((self.num_z, 1))))
            # 10. reassign lb, ub where possible
            A_ineq_i, b_ineq_i, A_eq_i, b_eq_i, lb_i, ub_i, z_map_constr_ineq_i, z_map_constr_eq_i = reassign_lb_ub_from_ineq(
                A_ineq_q, b_ineq_q, A_eq_q, b_eq_q, lb_q, ub_q, z_map_constr_ineq_q, z_map_constr_eq_q, z_map_vars_q)
            z_map_vars_i = z_map_vars_q
            # 11. objective: maximize distance between growth rate at production and no production
            c_i = c_p + [-c for c in c_b]

        # 3. Prepare module as undesired, desired or other
        if sd_module[MODULE_TYPE] == PROTECT:
            A_ineq_i, b_ineq_i, A_eq_i, b_eq_i, lb_i, ub_i, z_map_constr_ineq_i, z_map_constr_eq_i = reassign_lb_ub_from_ineq(
                A_ineq_p, b_ineq_p, A_eq_p, b_eq_p, lb_p, ub_p, z_map_constr_ineq_p, z_map_constr_eq_p, z_map_vars_p)
            z_map_vars_i = z_map_vars_p
            c_i = [0 for _ in range(A_ineq_i.shape[1])]
        elif sd_module[MODULE_TYPE] == SUPPRESS:
            A_ineq_i, b_ineq_i, A_eq_i, b_eq_i, lb_i, ub_i, z_map_constr_ineq_i, z_map_constr_eq_i, z_map_vars_i = farkas_dualize(
                A_ineq_p, b_ineq_p, A_eq_p, b_eq_p, lb_p, ub_p, z_map_constr_ineq_p, z_map_constr_eq_p, z_map_vars_p)
            c_i = [0 for _ in range(A_ineq_i.shape[1])]
        elif sd_module[MODULE_TYPE] == OPTKNOCK:
            A_ineq_i, b_ineq_i, A_eq_i, b_eq_i, lb_i, ub_i, z_map_constr_ineq_i, z_map_constr_eq_i = reassign_lb_ub_from_ineq(
                A_ineq_p, b_ineq_p, A_eq_p, b_eq_p, lb_p, ub_p, z_map_constr_ineq_p, z_map_constr_eq_p, z_map_vars_p)
            z_map_vars_i = z_map_vars_p
            # prepare outer objective
            c_i = linexprdict2mat(sd_module[OUTER_OBJECTIVE], self.model.reactions.list_attr('id'))
            c_i.resize((1, A_ineq_i.shape[1]))
            if OUTER_OPT_SENSE not in sd_module or sd_module[OUTER_OPT_SENSE] is None or \
                sd_module[OUTER_OPT_SENSE] not in [MINIMIZE, MAXIMIZE] or sd_module[OUTER_OPT_SENSE] == MAXIMIZE:
                c_i = -c_i
            c_i = c_i.toarray()[0].tolist()

        # 3. Add module to global MILP
        self.z_map_constr_ineq = sparse.hstack((self.z_map_constr_ineq, z_map_constr_ineq_i)).tocsc()
        self.z_map_constr_eq = sparse.hstack((self.z_map_constr_eq, z_map_constr_eq_i)).tocsc()
        self.z_map_vars = sparse.hstack((self.z_map_vars, z_map_vars_i)).tocsc()
        self.A_ineq = sparse.bmat([[self.A_ineq, None], [None, A_ineq_i]]).tocsr()
        self.b_ineq += b_ineq_i
        self.A_eq = sparse.bmat([[self.A_eq, None], [None, A_eq_i]]).tocsr()
        self.b_eq += b_eq_i
        self.c += c_i
        self.lb += lb_i
        self.ub += ub_i

    def link_z(self):
        """Connect binary intervention variables to variables and constraints of the strain design problem
        
        Function that uses the maps between intervention indicators z and variables and constraints
        of the linear strain design (in)equality system (self.z_map_constr_ineq, self.z_map_constr_eq and 
        self.z_map_vars) to set up the strain design MILP.
        
        MILP construction uses the following steps:
        
        (1) Translate equality-KOs/KIs to two inequality-KOs/KIs
        (2) Translate variable-KOs/KIs to inequality-KIs/KOs
        (3) Try to bound the problem with LPs
        (4) Use LP-determined bounds to link z-variables, where such bounds were found
        (5) Translate remaining inequalities back to equalities when possible and link z via indicator constraints.
        If necessary, the solver interface will translate them to big-M constraints.
        (6) Remove redundant equalities from static problem
        """

        # 1. Split knockable equality constraints into foward and reverse direction
        knockable_constr_eq = self.z_map_constr_eq.nonzero()[1]  # first array: z, second array: eq constr
        eq_constr_A = sparse.vstack((self.A_eq[knockable_constr_eq, :], -self.A_eq[knockable_constr_eq, :]))
        eq_constr_b = [self.b_eq[i] for i in knockable_constr_eq] + [-self.b_eq[i] for i in knockable_constr_eq]
        z_eq = self.z_map_constr_eq[:, tuple(knockable_constr_eq) * 2]
        # Add knockable inequalities to global A_ineq matrix
        self.A_ineq = sparse.vstack((self.A_ineq, eq_constr_A)).tocsr()
        self.b_ineq += eq_constr_b
        self.z_map_constr_ineq = sparse.hstack((self.z_map_constr_ineq, z_eq)).tocsc()
        # Remove knockable equalities from A_eq
        n_rows_eq = self.A_eq.shape[0]
        self.A_eq = self.A_eq[[False if i in knockable_constr_eq else True for i in range(0, n_rows_eq)]]
        self.b_eq = [self.b_eq[i] for i in range(0, len(self.b_eq)) if i not in knockable_constr_eq]
        self.z_map_constr_eq = self.z_map_constr_eq[:, [False if i in knockable_constr_eq else True for i in range(0, n_rows_eq)]]

        # 2. Translate all variable knockouts to inequality knockouts
        numvars = self.A_ineq.shape[1]
        knockable_vars = self.z_map_vars.nonzero()  # first array: z, second array: x
        knockable_vars_geq0 = [i for i in knockable_vars[1] if self.ub[i] > 0]
        knockable_vars_leq0 = [i for i in knockable_vars[1] if self.lb[i] < 0]
        ub_constr_A = sparse.csr_matrix(([1] * len(knockable_vars_geq0), (range(0, len(knockable_vars_geq0)), knockable_vars_geq0)),
                                        [len(knockable_vars_geq0), numvars])
        ub_constr_b = [0 for _ in knockable_vars_geq0]
        lb_constr_A = sparse.csr_matrix(([-1] * len(knockable_vars_leq0), (range(0, len(knockable_vars_leq0)), knockable_vars_leq0)),
                                        [len(knockable_vars_leq0), numvars])
        lb_constr_b = [0 for _ in knockable_vars_leq0]
        bnd_constr_A = sparse.vstack((ub_constr_A, lb_constr_A)).tocsr()
        bnd_constr_b = ub_constr_b + lb_constr_b
        var_kos = [knockable_vars[0][(knockable_vars[1] == i).nonzero()[0][0]] for i in knockable_vars_geq0 + knockable_vars_leq0]
        z_lb_ub = -self.z_map_vars[:, knockable_vars_geq0 + knockable_vars_leq0]
        # add constraints to main problem
        self.A_ineq = sparse.vstack((self.A_ineq, bnd_constr_A)).tocsr()
        self.b_ineq += bnd_constr_b
        self.z_map_constr_ineq = sparse.hstack((self.z_map_constr_ineq, z_lb_ub)).tocsc()

        # 3. Use LP to identify M-values for knockable constraints
        #    For this purpose, first construct a most relaxed LP-model (use all possible constraint-KOs, no possible var-KOs)
        knockable_constr_ineq = np.sort(self.z_map_constr_ineq.nonzero()[1])

        cont_vars = [False if i in self.idx_z else True for i in range(0, numvars)]
        M_A_ineq = self.A_ineq[[False if i in knockable_constr_ineq else True for i in range(0, self.A_ineq.shape[0])], :][:, cont_vars]
        M_b_ineq = [self.b_ineq[i] for i in range(0, self.A_ineq.shape[0]) if i not in knockable_constr_ineq]
        M_A_eq = self.A_eq[:, cont_vars]
        M_b_eq = self.b_eq.copy()
        M_lb = [self.lb[i] for i in np.nonzero(cont_vars)[0]]
        M_ub = [self.ub[i] for i in np.nonzero(cont_vars)[0]]
        # M_A contains a list of all knockable constraints. We need to maximize their value (M_A(i)*x) to get a good M
        # M_A(i)*x - z*M <= b
        # M*x <= b+z*M
        #     or
        # M_A(i)*x + z*M <= b + M
        # b is the right hand side value
        M_A = self.A_ineq[[True if i in knockable_constr_ineq else False for i in range(0, self.A_ineq.shape[0])], :][:, cont_vars]
        M_A = [(M_A[i, :])[0].toarray()[0] for i in range(M_A.shape[0])]
        M_b = [self.b_ineq[i] for i in range(0, self.A_ineq.shape[0]) if i in knockable_constr_ineq]

        processes = Configuration().processes
        num_Ms = len(M_A)
        processes = min(processes, num_Ms)

        max_Ax = [np.nan] * num_Ms

        # Dummy to check if optimization runs
        # worker_init(M_A,M_A_ineq,M_b_ineq,M_A_eq,M_b_eq,M_lb,M_ub,list(solvers.keys())[0])
        # worker_compute(1)

        logging.info('  Bounding MILP.')
        if processes > 1 and num_Ms > 1000:
            with SDPool(processes, initializer=worker_init,
                        initargs=(M_A, M_A_ineq, M_b_ineq, M_A_eq, M_b_eq, M_lb, M_ub, self.solver)) as pool:
                chunk_size = num_Ms // processes
                for i, value in pool.imap_unordered(worker_compute, range(num_Ms), chunksize=chunk_size):
                    max_Ax[i] = value
        else:
            worker_init(M_A, M_A_ineq, M_b_ineq, M_A_eq, M_b_eq, M_lb, M_ub, self.solver, self.seed)
            for i in range(num_Ms):
                _, max_Ax[i] = worker_compute(i)

        # round Ms up to 5 digits
        Ms = [np.ceil(M * 1e5) / 1e5 if not np.isinf(M) else self.M for M in max_Ax]
        # fill up M-vector also for notknockable reactions
        Ms = [
            Ms[np.array([i == j
                         for j in knockable_constr_ineq]).nonzero()[0][0]] if i in knockable_constr_ineq else np.nan
            for i in range(self.A_ineq.shape[0])
        ]

        # 4. Link constraints to z-variables for available upper bounds
        self.z_map_constr_ineq = self.z_map_constr_ineq.tocsc()
        self.A_ineq = self.A_ineq.todok()
        # iterate through knockable constraints
        for row in range(self.A_ineq.shape[0]):
            if not np.isinf(Ms[row]) and not np.isnan(Ms[row]):  # if there is a real number for M, use this for KO
                z_i = self.z_map_constr_ineq[:, row].nonzero()[0][0]
                sense = self.z_map_constr_ineq[z_i, row]
                if sense > 0:  # This means z_i = 1 knocks out ineq:
                    #     a_ineq*x - M*z <= b
                    self.A_ineq[row, z_i] = -Ms[row] + self.b_ineq[row]
                else:  # This means z_i = 0 knocks out ineq:
                    #     a_ineq*x + (M-b)*z <= M  (from a_ineq*x + M*z <= M + b)
                    self.A_ineq[row, z_i] = Ms[row] - self.b_ineq[row]
                    self.b_ineq[row] = Ms[row]
        self.z_map_constr_ineq = self.z_map_constr_ineq.tocsc()

        # 5. Translate back remaining inequalities to equations if applicable and link via indicator constraints
        knockable_constr_ineq = tuple(knockable_constr_ineq)
        knockable_constr_ineq_ic = [i for i in range(self.A_ineq.shape[0]) if np.isinf(Ms[i])]
        self.A_ineq = self.A_ineq.tocsr()

        # approach to find inequalities that can be lumped:
        # - construct a matrix from A_ineq, b_ineq, z_ineq for knockable constraints
        #   where every first entry of a row is positive
        # - search for row duplicates
        # - delete one if their first row entry had the same sign, lump to equality if they had opposite signs
        first_entry_A_ineq_sign = [np.sign(a.data[0]) if a.nnz > 0 else 0 for a in self.A_ineq]
        Ab_find_dupl = sparse.hstack((sparse.diags(first_entry_A_ineq_sign) * self.A_ineq, \
                                      sparse.coo_matrix(
                                          sparse.diags(first_entry_A_ineq_sign) * self.b_ineq).transpose(), \
                                      self.z_map_constr_ineq.transpose())) \
            .tocsr()
        # find rows that are identical
        ident_rows = []  # stores duplicate rows as Tuple (i,j,k): first row, second row, positive or negative duplicate
        for i, a1 in enumerate(Ab_find_dupl[range(Ab_find_dupl.shape[0] - 1)]):
            if i in knockable_constr_ineq_ic:  # only compare knockable ineqs
                for j, a2 in enumerate(Ab_find_dupl):
                    if j in knockable_constr_ineq_ic and j > i:  # only compare knockable ineqs
                        if a1.nnz == a2.nnz:
                            if all(a1.indices == a2.indices) & all(a1.data == a2.data):
                                ident_rows += [(i, j, first_entry_A_ineq_sign[i] * first_entry_A_ineq_sign[j])]
        # replace two ineqs by one eq
        self.z_map_constr_ineq = self.z_map_constr_ineq.tocsc()
        A_eq = sparse.csr_matrix((0, self.A_ineq.shape[1]))
        z_eq = sparse.csc_matrix((self.num_z, 0))
        b_eq = []
        for j in [i for i in range(len(ident_rows)) if ident_rows[i][2] == -1]:
            A_eq = sparse.vstack((A_eq, self.A_ineq[ident_rows[j][0]]))
            b_eq += [self.b_ineq[ident_rows[j][0]]]
            z_eq = sparse.hstack((z_eq, self.z_map_constr_ineq[:, ident_rows[j][0]]))
        # Add to global equality problem part
        knockable_constr_eq_ic = [i + len(self.b_eq) for i in range(len(b_eq))]
        self.A_eq = sparse.vstack((self.A_eq, A_eq), 'csr')
        self.b_eq += b_eq
        self.z_map_constr_eq = sparse.hstack((self.z_map_constr_eq, z_eq)).tocsc()

        # Remove all duplicates from ineq
        if ident_rows == []:
            remove_ineq = np.array([], 'int')
        else:
            remove_ineq = np.unique(np.hstack([[ir[0], ir[1]] if ir[2] == -1 else [ir[1]] for ir in ident_rows]))
        keep_ineq = [i for i in range(self.A_ineq.shape[0]) if i not in remove_ineq]
        knockable_constr_ineq_ic = [True if i in knockable_constr_ineq_ic else False for i in range(self.A_ineq.shape[0])]
        knockable_constr_ineq_ic = np.nonzero([knockable_constr_ineq_ic[i] for i in keep_ineq])[0]
        self.A_ineq = self.A_ineq[keep_ineq, :]
        self.b_ineq = [self.b_ineq[i] for i in keep_ineq]
        self.z_map_constr_ineq = self.z_map_constr_ineq[:, keep_ineq]

        # b5. Link remaining (in)equalities to z via indicator constraints
        #    and remove them from the static problem
        #    - first ineqs then eqs
        ic_binv = np.append(self.z_map_constr_ineq[:, knockable_constr_ineq_ic].indices,
                            self.z_map_constr_eq[:, knockable_constr_eq_ic].indices)
        ic_A = sparse.vstack((self.A_ineq[knockable_constr_ineq_ic, :], self.A_eq[knockable_constr_eq_ic, :]))
        ic_b = [self.b_ineq[i] for i in knockable_constr_ineq_ic] + [self.b_eq[i] for i in knockable_constr_eq_ic]
        ic_sense = 'L' * len(knockable_constr_ineq_ic) + 'E' * len(knockable_constr_eq_ic)
        ic_indicval = np.append(self.z_map_constr_ineq[:, knockable_constr_ineq_ic].data, self.z_map_constr_eq[:,
                                                                                                               knockable_constr_eq_ic].data)
        ic_indicval = [0 if i == 1 else 1 for i in ic_indicval]
        # in z-maps: -1 => z=1 -> A_ineq*x <= b_ineq
        #             1 => z=0 -> A_ineq*x <= b_ineq
        self.indic_constr = IndicatorConstraints(ic_binv, ic_A, ic_b, ic_sense, ic_indicval)

        # b6. Remove knockable (in)equalities from static problem, as they are now indicator constraints
        keep_ineq = [False if i in knockable_constr_ineq_ic else True for i in range(self.A_ineq.shape[0])]
        self.A_ineq = self.A_ineq[keep_ineq, :]
        self.b_ineq = [self.b_ineq[i] for i in range(len(keep_ineq)) if keep_ineq[i]]
        keep_eq = [False if i in knockable_constr_eq_ic else True for i in range(self.A_eq.shape[0])]
        self.A_eq = self.A_eq[keep_eq, :]
        self.b_eq = [self.b_eq[i] for i in range(len(keep_eq)) if keep_eq[i]]


class ContMILP:
    """Continuous representation of the strain design MILP.
    
    This MILP can be used to verify computation results. Since this class also stores the relationship
    between intervention variables z and corresponding (in)equality constraints and variables in the 
    problem, it can be used to verify computed designs quickly and in a numerically stable manner."""

    def __init__(self, A_ineq, b_ineq, A_eq, b_eq, lb, ub, c, z_map_constr_ineq, z_map_constr_eq, z_map_vars):
        self.A_ineq = A_ineq
        self.b_ineq = b_ineq
        self.A_eq = A_eq
        self.b_eq = b_eq
        self.lb = lb
        self.ub = ub
        self.c = c
        self.z_map_constr_ineq = z_map_constr_ineq
        self.z_map_constr_eq = z_map_constr_eq
        self.z_map_vars = z_map_vars

def build_primal_from_cbm(model, V_ineq=None, v_ineq=None, V_eq=None, v_eq=None, c=None) -> \
        Tuple[sparse.csr_matrix, Tuple, sparse.csr_matrix, Tuple, Tuple, Tuple, sparse.csr_matrix, sparse.csr_matrix, sparse.csr_matrix]:
    """Builds primal LP from constraint-based model and (optionally) additional constraints.
    
    standard form: A_ineq x <= b_ineq, A_eq x = b_eq, lb <= x <= ub, min{c'x}. Additionally,
    this function also returns a set of matrices that associate each variable (and constraint)
    with reactions. In the primal problems all variables correspond to reactions (z), therefore,
    the z_map_vars matrix is an identity matrix. The constraints correspond to metabolites, thus
    z_map_constr_ineq, z_map_constr_eq are all-zero.
    
    Args:
        model (cobra.Model):
            A metabolic model that is an instance of the cobra.Model class
            
        V_ineq, v_ineq (sparse.csr_matrix, list of float):
            Linear inequality constraints of the form V_ineq*x <= v_ineq. Ensure that the number
            of columns in V_ineq is identical to the number of reactions in the model.
            
        V_eq, v_eq (sparse.csr_matrix, list of float):
            Linear equality constraints of the form V_eq*x = v_eq. Ensure that the number
            of columns in V_eq is identical to the number of reactions in the model.
            
        c (list of float):
            Object coefficient vector (same lenght as variable vector).
    
    Returns:
        (Tuple):
        A_ineq, b_ineq, A_eq, b_eq, lb, ub, c, z_map_constr_ineq, z_map_constr_eq, z_map_vars.
        A constraint-based steady-state model in the form of a linear (in)equality system. The matrices
        z_map_constr_ineq, z_map_constr_eq, z_map_vars contain the association between reactions and
        different parts of the LP, such as reactions, metabolites or other (in)equalities.
    """
    numr = len(model.reactions)
    # initialize matices (if not provided in function call)
    if V_ineq is None or v_ineq is None:
        V_ineq = sparse.csr_matrix((0, numr))
        v_ineq = []
    if V_eq is None or v_eq is None:
        V_eq = sparse.csr_matrix((0, numr))
        v_eq = []
    if c is None:
        c = [i.objective_coefficient for i in model.reactions]
    S = sparse.csr_matrix(create_stoichiometric_matrix(model))
    # fill matrices
    A_eq = sparse.vstack((S, V_eq))
    b_eq = [0 for _ in range(S.shape[0])] + v_eq
    A_ineq = V_ineq.copy()
    b_ineq = v_ineq.copy()
    lb = [v.lower_bound for v in model.reactions]
    ub = [v.upper_bound for v in model.reactions]
    z_map_vars = sparse.identity(numr, 'd', format="csc")
    z_map_constr_eq = sparse.csc_matrix((numr, A_eq.shape[0]))
    z_map_constr_ineq = sparse.csc_matrix((numr, A_ineq.shape[0]))
    A_ineq, b_ineq, lb, ub, z_map_constr_ineq = prevent_boundary_knockouts(A_ineq, b_ineq, lb.copy(), ub.copy(), z_map_constr_ineq,
                                                                           z_map_vars)
    return A_ineq, b_ineq, A_eq, b_eq, lb, ub, c, z_map_constr_ineq, z_map_constr_eq, z_map_vars

def LP_dualize(A_ineq_p, b_ineq_p, A_eq_p, b_eq_p, lb_p, ub_p, c_p,
            z_map_constr_ineq_p=None, z_map_constr_eq_p=None, z_map_vars_p=None) -> \
        Tuple[sparse.csr_matrix, Tuple, sparse.csr_matrix, Tuple, Tuple, Tuple, sparse.csr_matrix, sparse.csr_matrix, sparse.csr_matrix]:
    """Translate a primal system to its LP dual system
    
    The primal system must be given in the standard form: A_ineq x <= b_ineq, A_eq x = b_eq, lb <= x < ub, min{c'x}. The LP duality
    theorem defines a set of two problems. If one of the LPs is a maximization and and optimum exists, the optimal value of this LP
    is identical to the minimal optimum of its LP dual problem. LP duality can be used for nested optimization, since solving the 
    primal and the LP dual problem, while enfocing equality of the objective value, guarantees optimality.
    
    Construction of the LP dual:    
        Variables translate to constraints:
            x={R} ->   =
            x>=0  ->  >= (new constraint is multiplied with -1 to translate to <= e.g. -A_i' y <= -c_i)
            x<=0  ->  <=
        Constraints translate to variables:
            =     ->   y={R}
            <=    ->   y>=0
            
    Args:
        A_ineq_p, b_ineq_p (sparse.csr_matrix and list of float): 
            A coefficient matrix and a vector that describe the linear inequalities of the primal LP A_ineq_p*x <= b_ineq_p
            
        A_eq_p, b_eq_p (sparse.csr_matrix and list of float): 
            A coefficient matrix and a vector that describe the linear equalities of the primal LP A_eq_p*x <= b_eq_p
            
        lb_p, ub_p (list of float):
            Upper and lower variable bounds in vector form.
            
        c_p (list of float):
            The objective coefficient vector of the primal minimization-LP.
            z_map_constr_ineq_p, z_map_constr_eq_p, z_map_vars_p
                    
        z_map_constr_ineq, z_map_constr_eq, z_map_vars (optional (sparse.csr_matrix)):
            Matrices that contain the relationship between metabolic reactions and different parts of the LP, such as reactions, 
            metabolites or other (in)equalities. These matrices help keeping track of the parts of the LP that are affected by
            reaction knockouts and additions. When a reaction (i) knockout removes the variable or constraint (j), the respective
            matrix contains a coefficient 1 at this position. -1 marks additions. E.g.: If the knockout of reaction i corresponds
            to the removal of inequality constraint j, there is a matrix entry z_map_constr_ineq_(i,j) = 1. If these matrices are
            provided, they are updated for the dualized LP, if not, the dual problem is constructed without returning information 
            about these relationships.
            
    Returns
        (Tuple):
        The LP dual of the problem in the format: A_ineq, b_ineq, A_eq, b_eq, c, lb, ub
        and optionally also z_map_constr_ineq, z_map_constr_eq, z_map_vars
    """
    numr = A_ineq_p.shape[1]
    numz = max([z.shape[0] for z in [z_map_vars_p, z_map_constr_eq_p, z_map_constr_ineq_p] if z is not None])

    if numz:
        if z_map_vars_p is None:
            z_map_vars_p = sparse.csc_matrix((numz, A_ineq_p.shape[1]))
        if z_map_constr_eq_p is None:
            z_map_constr_eq_p = sparse.csc_matrix((numz, A_eq_p.shape[0]))
        if z_map_constr_ineq_p is None:
            z_map_constr_ineq_p = sparse.csc_matrix((numz, A_ineq_p.shape[0]))

    # knockouts of variables and constraints must not overlap in the problem matrix
    if not len(
            A_eq_p[[True if i in z_map_constr_eq_p.nonzero()[1] else False for i in range(0, A_eq_p.shape[0])], :] \
                    [:,
            [True if i in z_map_vars_p.nonzero()[1] else False for i in range(0, A_eq_p.shape[1])]].nonzero()[
                0]) == 0 \
            or not len(
        A_ineq_p[[True if i in z_map_constr_ineq_p.nonzero()[1] else False for i in range(0, A_ineq_p.shape[0])], :] \
                [:,
        [True if i in z_map_vars_p.nonzero()[1] else False for i in range(0, A_ineq_p.shape[1])]].nonzero()[
            0]) == 0:
        raise Exception(
            "knockouts of variables and constraints must not overlap in the problem matrix. Something went wrong during the construction of the primal problem."
        )

    if c_p == []:
        c_p = [0.0] * numr

    # Translate inhomogenous bounds into inequality constraints
    lb_inh_bounds = [i for i in np.nonzero(lb_p)[0] if not np.isinf(lb_p[i])]
    ub_inh_bounds = [i for i in np.nonzero(ub_p)[0] if not np.isinf(ub_p[i])]
    x_geq0 = np.nonzero(np.greater_equal(lb_p, 0) & np.greater(ub_p, 0))[0]
    x_eR = np.nonzero(np.greater(0, lb_p) & np.greater(ub_p, 0))[0]
    x_leq0 = np.nonzero(np.greater(0, lb_p) & np.greater_equal(0, ub_p))[0]

    LB = sparse.csr_matrix((len(lb_inh_bounds) * [-1], (range(0, len(lb_inh_bounds)), lb_inh_bounds)), shape=(len(lb_inh_bounds), numr))
    UB = sparse.csr_matrix((len(ub_inh_bounds) * [1], (range(0, len(ub_inh_bounds)), ub_inh_bounds)), shape=(len(ub_inh_bounds), numr))
    A_ineq_p = sparse.vstack((A_ineq_p, LB, UB))
    b_ineq_p = b_ineq_p + [-lb_p[i] for i in lb_inh_bounds] + [ub_p[i] for i in ub_inh_bounds]

    # Translate into dual system
    # Transpose parts of matrices with variables >=0,<=0 to put them into A_ineq
    A_ineq = sparse.vstack((sparse.hstack((np.transpose(-A_eq_p[:, x_geq0]), np.transpose(-A_ineq_p[:, x_geq0]))), \
                            sparse.hstack(
                                (np.transpose(A_eq_p[:, x_leq0]), np.transpose(A_ineq_p[:, x_leq0]))))).tocsr()
    b_ineq = [c_p[i] for i in x_geq0] + [-c_p[i] for i in x_leq0]
    A_eq = sparse.hstack((np.transpose(A_eq_p[:, x_eR]), np.transpose(A_ineq_p[:, x_eR]))).tocsr()
    b_eq = [c_p[i] for i in x_eR]
    lb = [-np.inf] * A_eq_p.shape[0] + [0.0] * A_ineq_p.shape[0]
    ub = [np.inf] * (A_eq_p.shape[0] + A_ineq_p.shape[0])
    c = b_eq_p + b_ineq_p

    if numz:
        # translate mapping of z-variables to rows instead of columns
        z_map_constr_ineq = sparse.hstack((z_map_vars_p[:, x_geq0], z_map_vars_p[:, x_leq0])).tocsc()
        z_map_constr_eq = z_map_vars_p[:, x_eR]
        z_map_vars = sparse.hstack(
            (z_map_constr_eq_p, z_map_constr_ineq_p, sparse.csc_matrix((numz, len(lb_inh_bounds) + len(ub_inh_bounds))))).tocsc()
        A_ineq, b_ineq, A_eq, b_eq, lb, ub, z_map_constr_ineq, z_map_constr_eq = \
            reassign_lb_ub_from_ineq(A_ineq, b_ineq, A_eq, b_eq, lb, ub,  z_map_constr_ineq, z_map_constr_eq, z_map_vars)
        return A_ineq, b_ineq, A_eq, b_eq, lb, ub, c, z_map_constr_ineq, z_map_constr_eq, z_map_vars
    else:
        A_ineq, b_ineq, A_eq, b_eq, lb, ub = reassign_lb_ub_from_ineq(A_ineq, b_ineq, A_eq, b_eq, lb, ub)
        return A_ineq, b_ineq, A_eq, b_eq, lb, ub, c

def farkas_dualize(A_ineq_p, b_ineq_p, A_eq_p, b_eq_p, lb_p, ub_p,
                  z_map_constr_ineq_p=None, z_map_constr_eq_p=None, z_map_vars_p=None) -> \
        Tuple[sparse.csr_matrix, Tuple, sparse.csr_matrix, Tuple, Tuple, sparse.csr_matrix, sparse.csr_matrix, sparse.csr_matrix]:
    """Translate a primal system of linear (in)equality to its Farkas dual
    
    The primal system must be given in the standard form: A_ineq x <= b_ineq, A_eq x = b_eq, lb <= x < ub. Farkas' lemma
    defines a set of two systems of linear (in)equalities of which exactly one is feasible. Since the feasibility of one is a
    certificate for the infeasibility of the other one, this theorem can be used to set up problems that imply the infeasibility
    and thus exclusion of a certain subspace. This priciple is used for MCS calculation (the SUPPRESS module).
    
    Consider that the following is not implemented:
    In the case of (1) A x = b, (2) x={R}, (3) b~=0, Farkas' lemma is special,
    because b'y ~= 0 is required to make the primal infeasible instead of b'y < 0.
    1. This does not occur very often.
    2. Splitting the equality into two inequalities that translate to y>=0
    would be posible, and yield b'y < 0 in the farkas' lemma.
    Maybe splitting is required, but I actually don't think so. Using the
    special case of b'y < 0 for b'y ~= 0 should be enough.
    
    Args:
        A_ineq_p, b_ineq_p (sparse.csr_matrix and list of float): 
            A coefficient matrix and a vector that describe the linear inequalities of the primal LP A_ineq_p*x <= b_ineq_p
            
        A_eq_p, b_eq_p (sparse.csr_matrix and list of float): 
            A coefficient matrix and a vector that describe the linear equalities of the primal LP A_eq_p*x <= b_eq_p
            
        lb_p, ub_p (list of float):
            Upper and lower variable bounds in vector form.
                    
        z_map_constr_ineq, z_map_constr_eq, z_map_vars (optional (sparse.csr_matrix)):
            Matrices that contain the relationship between metabolic reactions and different parts of the LP, such as reactions, 
            metabolites or other (in)equalities. These matrices help keeping track of the parts of the LP that are affected by
            reaction knockouts and additions. When a reaction (i) knockout removes the variable or constraint (j), the respective
            matrix contains a coefficient 1 at this position. -1 marks additions. E.g.: If the knockout of reaction i corresponds
            to the removal of inequality constraint j, there is a matrix entry z_map_constr_ineq_(i,j) = 1. If these matrices are
            provided, they are updated for the dualized LP, if not, the dual problem is constructed without returning information 
            about these relationships.
            
    Returns
        (Tuple):
        The Farkas dual of the linear (in)equality system in the format: A_ineq, b_ineq, A_eq, b_eq, lb, ub
        and optionally also z_map_constr_ineq, z_map_constr_eq, z_map_vars
    """
    numz = max([0] + [z.shape[0] for z in [z_map_vars_p, z_map_constr_eq_p, z_map_constr_ineq_p] if z is not None])
    c_p = [0 for _ in range(A_ineq_p.shape[1])]

    if numz:
        A_ineq_d, b_ineq_d, A_eq_d, b_eq_d, lb_f, ub_f, c_d, z_map_constr_ineq_d, z_map_constr_eq_f, z_map_vars_f = LP_dualize(
            A_ineq_p, b_ineq_p, A_eq_p, b_eq_p, lb_p, ub_p, c_p, z_map_constr_ineq_p, z_map_constr_eq_p, z_map_vars_p)
    else:
        A_ineq_d, b_ineq_d, A_eq_d, b_eq_d, lb_f, ub_f, c_d = LP_dualize(A_ineq_p, b_ineq_p, A_eq_p, b_eq_p, lb_p, ub_p, c_p)
    # add constraint b_prim'y or (c_dual'*y) <= -1;
    A_ineq_f = sparse.vstack((A_ineq_d, sparse.csr_matrix(c_d))).tocsr()
    b_ineq_f = b_ineq_d + [-1]
    A_eq_f = A_eq_d
    b_eq_f = b_eq_d
    # it would also be possible (but ofc not necessary) to force (c_dual*y) == -1; instead
    # A_eq = sparse.vstack((A_eq,c_dual)).tocsr()
    # b_eq += [-1]
    if numz:
        z_map_constr_ineq_f = sparse.hstack((z_map_constr_ineq_d, sparse.csr_matrix((numz, 1)))).tocsc()
        # z_map_constr_eq_f = sparse.hstack((z_map_constr_eq_d,sparse.csr_matrix((numz,1)))).tocsc()
        return A_ineq_f, b_ineq_f, A_eq_f, b_eq_f, lb_f, ub_f, z_map_constr_ineq_f, z_map_constr_eq_f, z_map_vars_f
    else:
        return A_ineq_f, b_ineq_f, A_eq_f, b_eq_f, lb_f, ub_f, z_map_constr_ineq_f, z_map_constr_eq_f, z_map_vars_f

def reassign_lb_ub_from_ineq(A_ineq, b_ineq, A_eq, b_eq, lb, ub,
                             z_map_constr_ineq=None, z_map_constr_eq=None, z_map_vars=None) -> \
        Tuple[sparse.csr_matrix, Tuple, sparse.csr_matrix, Tuple, Tuple, Tuple, sparse.csr_matrix, sparse.csr_matrix]:
    """Remove single constraints in A_ineq or A_eq in favor of lower and upper bounds on variables
    
    Constraints on single variables instead translated into lower and upper bounds (lb, ub). This
    is useful to filter out redundant bounds on variables and keep the (in)equality system concise. 
    To avoid interference with the knock-out logic, negative upper bounds and positive lower bounds 
    are not put into lb and ub, when reactions are flagged knockable with z_map_vars.
    
    Args:
        A_ineq_p, b_ineq_p (sparse.csr_matrix and list of float): 
            A coefficient matrix and a vector that describe the linear inequalities of the primal LP A_ineq_p*x <= b_ineq_p
            
        A_eq_p, b_eq_p (sparse.csr_matrix and list of float): 
            A coefficient matrix and a vector that describe the linear equalities of the primal LP A_eq_p*x <= b_eq_p
            
        lb_p, ub_p (list of float):
            Upper and lower variable bounds in vector form.
            
        z_map_constr_ineq, z_map_constr_eq, z_map_vars (optional (sparse.csr_matrix)):
            Matrices that contain the relationship between metabolic reactions and different parts of the LP, such as reactions, 
            metabolites or other (in)equalities. These matrices help keeping track of the parts of the LP that are affected by
            reaction knockouts and additions. When a reaction (i) knockout removes the variable or constraint (j), the respective
            matrix contains a coefficient 1 at this position. -1 marks additions. E.g.: If the knockout of reaction i corresponds
            to the removal of inequality constraint j, there is a matrix entry z_map_constr_ineq_(i,j) = 1. If these matrices are
            provided, they are updated. Otherwise, all reactions are assumed to be notknockable and thus all constraints on 
            single variables put into lb and ub.
    
    Returns
        (Tuple):
        A linear (in)equality system in the format: A_ineq, b_ineq, A_eq, b_eq, lb, ub
        and optionally also updated z_map_constr_ineq, z_map_constr_eq
    """
    lb = [[l] for l in lb]
    ub = [[u] for u in ub]
    numz = max([0] + [z.shape[0] for z in [z_map_vars, z_map_constr_eq, z_map_constr_ineq] if z is not None])
    numr = A_ineq.shape[1]

    if z_map_vars is None:
        z_map_vars = sparse.csc_matrix((numz, numr))

    # translate entries to lb or ub
    # find all entries in A_ineq
    row_ineq = A_ineq.nonzero()[0]
    # filter for rows with only one entry
    var_bound_constraint_ineq = [i for i in row_ineq if list(row_ineq).count(i) == 1]
    # exclude knockable constraints
    var_bound_constraint_ineq = [i for i in var_bound_constraint_ineq if i not in z_map_constr_ineq.nonzero()[1]]
    # retrieve all bounds from inequality constraints
    for i in var_bound_constraint_ineq:
        idx_r = A_ineq[i, :].nonzero()[1][0]  # get reaction from constraint (column of entry)
        if A_ineq[i, idx_r] > 0:  # upper bound constraint
            ub[idx_r] += [b_ineq[i] / A_ineq[i, idx_r]]
        else:  # lower bound constraint
            lb[idx_r] += [b_ineq[i] / A_ineq[i, idx_r]]

    # find all entries in A_eq
    row_eq = A_eq.nonzero()[0]
    # filter for rows with only one entry
    var_bound_constraint_eq = [i for i in row_eq if list(row_eq).count(i) == 1]
    # exclude knockable constraints
    var_bound_constraint_eq = [i for i in var_bound_constraint_eq if i not in z_map_constr_eq.nonzero()[1]]
    # retrieve all bounds from equality constraints
    # and partly set lb or ub derived from equality constraints, for instance:
    # If x =  5, set ub = 5 and keep the inequality constraint -x <= -5.
    # If x = -5, set lb =-5 and keep the inequality constraint  x <= -5.
    A_ineq_new = sparse.csr_matrix((0, numr))
    b_ineq_new = []
    for i in var_bound_constraint_eq:
        idx_r = A_eq[i, :].nonzero()[1][0]  # get reaction from constraint (column of entry)
        if any(z_map_vars[:, idx_r]):  # if reaction is knockable
            if A_eq[i, idx_r] * b_eq[i] > 0:  # upper bound constraint
                ub[idx_r] += [b_eq[i] / A_eq[i, idx_r]]
                A_ineq_new = sparse.vstack((A_ineq_new, -A_eq[i, :]))
                b_ineq_new += [-b_eq[i]]
            elif A_eq[i, idx_r] * b_eq[i] < 0:  # lower bound constraint
                lb[idx_r] += [b_eq[i] / A_eq[i, idx_r]]
                A_ineq_new = sparse.vstack((A_ineq_new, A_eq[i, :]))
                b_ineq_new += [b_eq[i]]
            else:
                ub[idx_r] += [0.0]
                lb[idx_r] += [0.0]
        else:
            lb[idx_r] += [b_eq[i] / A_eq[i, idx_r]]
            ub[idx_r] += [b_eq[i] / A_eq[i, idx_r]]
    # set tightest bounds (avoid inf)
    lb = [max([i for i in l if not np.isinf(i)] + [np.nan]) for l in lb]
    ub = [min([i for i in u if not np.isinf(i)] + [np.nan]) for u in ub]
    # set if only if no other bound remains
    lb = [-np.inf if np.isnan(l) else l for l in lb]
    ub = [np.inf if np.isnan(u) else u for u in ub]

    # check if bounds are consistent
    if any(np.greater(lb, ub)):
        raise Exception("There is a lower bound that is greater than its upper bound counterpart.")

    # remove constraints that became redundant
    numineq = A_ineq.shape[0]
    A_ineq = A_ineq[[False if i in var_bound_constraint_ineq else True for i in range(0, numineq)]]
    b_ineq = [b_ineq[i] for i in range(0, len(b_ineq)) if i not in var_bound_constraint_ineq]
    z_map_constr_ineq = z_map_constr_ineq[:, [False if i in var_bound_constraint_ineq else True for i in range(0, numineq)]]
    numeq = A_eq.shape[0]
    A_eq = A_eq[[False if i in var_bound_constraint_eq else True for i in range(0, numeq)]]
    b_eq = [b_eq[i] for i in range(0, len(b_eq)) if i not in var_bound_constraint_eq]
    # add equality constraints that transformed to inequality constraints
    A_ineq = sparse.vstack((A_ineq, A_ineq_new))
    b_ineq += b_ineq_new
    if numz:
        z_map_constr_eq = z_map_constr_eq[:, [False if i in var_bound_constraint_eq else True for i in range(0, numeq)]]
        z_map_constr_ineq = sparse.hstack((z_map_constr_ineq, sparse.csc_matrix((numz, A_ineq_new.shape[0]))))
        return A_ineq, b_ineq, A_eq, b_eq, lb, ub, z_map_constr_ineq, z_map_constr_eq
    else:
        return A_ineq, b_ineq, A_eq, b_eq, lb, ub

def prevent_boundary_knockouts(A_ineq, b_ineq, lb, ub, z_map_constr_ineq, z_map_vars) -> \
        Tuple[sparse.csr_matrix, Tuple, Tuple, Tuple, sparse.csr_matrix]:
    """Put negative lower bounds and positive upper bounds into (notknockable) inequalities
    
    This is a helper function that puts negative lower bounds and positive upper bounds into (not-knockable) 
    inequalities. Later on, one may simulate the knockouts by multiplying the upper and lower bounds with a
    binary variable z. This functions prevents that 
    
    Args:
        A_ineq_p, b_ineq_p (sparse.csr_matrix and list of float): 
            A coefficient matrix and a vector that describe the linear inequalities of the primal LP A_ineq_p*x <= b_ineq_p
            
        lb_p, ub_p (list of float):
            Upper and lower variable bounds in vector form.
            
        z_map_constr_ineq, z_map_vars (optional (sparse.csr_matrix)):
            Matrices that contain the relationship between metabolic reactions and different parts of the LP, such as reactions, 
            metabolites or other (in)equalities. These matrices help keeping track of the parts of the LP that are affected by
            reaction knockouts and additions. When a reaction (i) knockout removes the variable or constraint (j), the respective
            matrix contains a coefficient 1 at this position. -1 marks additions. E.g.: If the knockout of reaction i corresponds
            to the removal of inequality constraint j, there is a matrix entry z_map_constr_ineq_(i,j) = 1. If these matrices are
            provided, they are updated. Otherwise, all reactions are assumed to be knockable and thus all negative upper and 
            positive lower bounds are translated into constraints.           
    
    Returns
        (Tuple):
        A linear (in)equality system in the format: A_ineq, b_ineq, A_eq, b_eq, lb, ub
        and optionally also updated z_map_constr_ineq, z_map_constr_eq
    """
    numr = A_ineq.shape[1]
    numz = max([0] + [z.shape[0] for z in [z_map_vars, z_map_constr_ineq] if z is not None])

    if z_map_vars is None:
        z_map_vars = sparse.csc_matrix((numz, numr))

    for i in range(0, numr):
        if any(z_map_vars[:, i]) and lb[i] > 0:
            A_ineq = sparse.vstack((A_ineq, sparse.csr_matrix(([-1], ([0], [i])), shape=(1, numr))))
            b_ineq += [-lb[i]]
            z_map_constr_ineq = sparse.hstack((z_map_constr_ineq, sparse.csc_matrix((numz, 1))))
            lb[i] = 0.0
        if any(z_map_vars[:, i]) and ub[i] < 0:
            A_ineq = sparse.vstack((A_ineq, sparse.csr_matrix(([1], ([0], [i])), shape=(1, numr))))
            b_ineq += [ub[i]]
            z_map_constr_ineq = sparse.hstack((z_map_constr_ineq, sparse.csc_matrix((numz, 1))))
            ub[i] = 0.0

    return A_ineq, b_ineq, lb, ub, z_map_constr_ineq


def worker_init(A, A_ineq, b_ineq, A_eq, b_eq, lb, ub, solver, seed):
    """Helper function for determining bounds on linear expressions"""
    global lp_glob
    lp_glob = MILP_LP(A_ineq=A_ineq, b_ineq=b_ineq, A_eq=A_eq, b_eq=b_eq, lb=lb, ub=ub, solver=solver, seed=seed)
    if lp_glob == CPLEX:
        lp_glob.backend.parameters.lpmethod.set(1)
        if Configuration().processes > 1:
            lp_glob.backend.parameters.threads.set(2)
    lp_glob.solver = solver
    lp_glob.A = A


def worker_compute(i) -> Tuple[int, float]:
    """Helper function for determining bounds on linear expressions"""
    global lp_glob
    # maximize by minimizing negative objective and negating result
    lp_glob.set_objective(-lp_glob.A[i])
    min_cx = -lp_glob.slim_solve()
    return i, min_cx



#!/usr/bin/env python3
#
# Copyright 2022 Max Planck Insitute Magdeburg
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
#
"""Functions for the compression of metabolic networks, taken from the efmtool_link package

Functions in this module not meant to be used outside the compression of networks. For a
the documentation of the efmtool compression provided by StrainDesign, refer to the networktools
module."""

import numpy
import jpype
import os
import subprocess
import sympy
import io
from contextlib import redirect_stdout, redirect_stderr
"""Initialization of the java machine, since efmtool compression is done in java."""
efmtool_jar = os.path.join(os.path.dirname(__file__), 'efmtool.jar')
jpype.addClassPath(efmtool_jar)
if not jpype.isJVMStarted():
    with redirect_stdout(io.StringIO()), redirect_stderr(io.StringIO()):  # suppress console output
        # mem_bytes = os.sysconf('SC_PAGE_SIZE') * os.sysconf('SC_PHYS_PAGES')  # e.g. 4015976448
        # mem_mb = round(mem_bytes/(1024.**2)*0.75) # allow 75% of total memory for heap space
        # mem_mb = round(psutil.virtual_memory()[0]/(1024.**2)*0.75)
        # jpype.startJVM( jpype.getDefaultJVMPath() , f"-Xmx{mem_mb}m" )
        jpype.startJVM()
import jpype.imports

import ch.javasoft.smx.impl.DefaultBigIntegerRationalMatrix as DefaultBigIntegerRationalMatrix
import ch.javasoft.smx.ops.Gauss as Gauss
import ch.javasoft.metabolic.compress.CompressionMethod as CompressionMethod

subset_compression = CompressionMethod[:](
    [CompressionMethod.CoupledZero, CompressionMethod.CoupledCombine, CompressionMethod.CoupledContradicting])
import ch.javasoft.metabolic.compress.StoichMatrixCompressor as StoichMatrixCompressor
import ch.javasoft.math.BigFraction as BigFraction
import java.math.BigInteger as BigInteger

jTrue = jpype.JBoolean(True)
jSystem = jpype.JClass("java.lang.System")

# try to find a working java executable
_java_executable = 'java'
try:
    cp = subprocess.run([_java_executable, '-version'], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    if cp.returncode != 0:
        _java_executable = ''
except:
    _java_executable = ''
if _java_executable == '':
    _java_executable = os.path.join(os.environ.get('JAVA_HOME', ''), "bin", "java")
    try:
        cp = subprocess.run([_java_executable, '-version'], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        if cp.returncode != 0:
            _java_executable = ''
    except:
        _java_executable = ''
if _java_executable == '':
    import efmtool_link.efmtool_intern  # just to find java executable via jpype
    _java_executable = os.path.join(str(efmtool_link.efmtool_intern.jSystem.getProperty("java.home")), "bin", "java")


def basic_columns_rat(mx, tolerance=0):  # mx is ch.javasoft.smx.impl.DefaultBigIntegerRationalMatrix
    """efmtool: Translate matrix coefficients to rational numbers"""
    if type(mx) is numpy.ndarray:
        mx = DefaultBigIntegerRationalMatrix(numpy_mat2jpypeArrayOfArrays(mx), jTrue, jTrue)
    row_map = jpype.JInt[mx.getRowCount()]  # just a placeholder because we don't care about the row permutation here
    col_map = jpype.JInt[:](range(mx.getColumnCount()))
    rank = Gauss.getRationalInstance().rowEchelon(mx, False, row_map, col_map)

    return col_map[0:rank]


def numpy_mat2jpypeArrayOfArrays(npmat):
    """efmtool: Translate matrix to array of arrays"""
    rows = npmat.shape[0]
    cols = npmat.shape[1]
    jmat = jpype.JDouble[rows, cols]
    # for sparse matrices can use nonzero() here instead of iterating through everything
    for r in range(rows):
        for c in range(cols):
            jmat[r][c] = npmat[r, c]
    return jmat


def jpypeArrayOfArrays2numpy_mat(jmat):
    """efmtool: Translate array of arrays to numpy matrix"""
    rows = len(jmat)
    cols = len(jmat[0])  # assumes all rows have the same number of columns
    npmat = numpy.zeros((rows, cols))
    for r in range(rows):
        for c in range(cols):
            npmat[r, c] = jmat[r][c]
    return npmat


def sympyRat2jBigIntegerPair(val):
    """efmtool: Translate rational numbers to big integer pair"""
    numer = val.p  # numerator
    if numer.bit_length() <= 63:
        numer = BigInteger.valueOf(numer)
    else:
        numer = BigInteger(str(numer))
    denom = val.q  # denominator
    if denom.bit_length() <= 63:
        denom = BigInteger.valueOf(denom)
    else:
        denom = BigInteger(str(denom))
    return (numer, denom)


def jBigFraction2sympyRat(val):
    """efmtool: Translate rational numbers to sympy rational numbers"""
    return jBigIntegerPair2sympyRat(val.getNumerator(), val.getDenominator())


def jBigIntegerPair2sympyRat(numer, denom):
    """efmtool: Translate big integer pair to sympy rational numbers"""
    if numer.bitLength() <= 63:
        numer = numer.longValue()
    else:
        numer = str(numer.toString())
    if denom.bitLength() <= 63:
        denom = denom.longValue()
    else:
        denom = str(denom.toString())
    return sympy.Rational(numer, denom)
